# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division
from blockdiag.noderenderer import install_renderer
from blockdiag.noderenderer.base import NodeShape
from blockdiag.utils import XY, Box, Size


class Actor(NodeShape):
    def __init__(self, node, metrics=None):
        super(Actor, self).__init__(node, metrics)

        m = metrics.cell(node)
        if node.label:
            font = metrics.font_for(self.node)
            textsize = metrics.textsize(node.label, font)
            shortside = min(m.width, m.height - textsize.height)
        else:
            textsize = Size(0, 0)
            shortside = min(m.width, m.height)

        r = self.radius = shortside // 8  # radius of actor's head
        self.center = metrics.cell(node).center

        self.connectors[0] = XY(self.center.x, self.center.y - r * 9 // 2)
        self.connectors[1] = XY(self.center.x + r * 4, self.center.y)
        self.connectors[2] = XY(self.center.x,
                                self.center.y + r * 4 + textsize.height)
        self.connectors[3] = XY(self.center.x - r * 4, self.center.y)

        self.textbox = Box(m.left.x,
                           self.center.y + r * 4,
                           m.right.x,
                           self.connectors[2].y)

    def head_part(self):
        r = self.radius * 3 // 2
        pt = self.metrics.cell(self.node).center.shift(y=-self.radius * 3)
        return Box(pt.x - r, pt.y - r, pt.x + r, pt.y + r)

    def body_part(self):
        r = self.radius
        m = self.metrics.cell(self.node)

        bodyC = m.center
        neckWidth = r * 2 // 3  # neck size
        arm = r * 4  # arm length
        armWidth = r
        bodyWidth = r * 2 // 3  # half of body width
        bodyHeight = r
        legXout = r * 7 // 2  # toe outer position
        legYout = bodyHeight + r * 3
        legXin = r * 2  # toe inner position
        legYin = bodyHeight + r * 3

        return [XY(bodyC.x + neckWidth, bodyC.y - r * 2),
                XY(bodyC.x + neckWidth, bodyC.y - armWidth),  # neck end
                XY(bodyC.x + arm, bodyC.y - armWidth),
                XY(bodyC.x + arm, bodyC.y),  # right arm end
                XY(bodyC.x + bodyWidth, bodyC.y),   # right body end
                XY(bodyC.x + bodyWidth, bodyC.y + bodyHeight),
                XY(bodyC.x + legXout, bodyC.y + legYout),
                XY(bodyC.x + legXin, bodyC.y + legYin),

                XY(bodyC.x, bodyC.y + (bodyHeight * 2)),  # body bottom center

                XY(bodyC.x - legXin, bodyC.y + legYin),
                XY(bodyC.x - legXout, bodyC.y + legYout),
                XY(bodyC.x - bodyWidth, bodyC.y + bodyHeight),
                XY(bodyC.x - bodyWidth, bodyC.y),  # left body end
                XY(bodyC.x - arm, bodyC.y),
                XY(bodyC.x - arm, bodyC.y - armWidth),
                XY(bodyC.x - neckWidth, bodyC.y - armWidth),  # left arm end
                XY(bodyC.x - neckWidth, bodyC.y - r * 2)]

    def render_shape(self, drawer, _, **kwargs):
        fill = kwargs.get('fill')

        # draw body part
        body = self.body_part()
        if kwargs.get('shadow'):
            body = self.shift_shadow(body)
            if kwargs.get('style') == 'blur':
                drawer.polygon(body, fill=fill, filter='transp-blur')
            else:
                drawer.polygon(body, fill=fill)
        else:
            drawer.polygon(body, fill=self.node.color,
                           outline=self.node.linecolor, style=self.node.style)

        # draw head part
        head = self.head_part()
        if kwargs.get('shadow'):
            head = self.shift_shadow(head)
            if kwargs.get('style') == 'blur':
                drawer.ellipse(head, fill=fill, outline=self.node.linecolor,
                               filter='transp-blur')
            else:
                drawer.ellipse(head, fill=fill, outline=self.node.linecolor)
        else:
            drawer.ellipse(head, fill=self.node.color,
                           outline=self.node.linecolor, style=self.node.style)


def setup(self):
    install_renderer('actor', Actor)



# -*- coding: utf-8 -*-

# Copyright (c) 2008/2009 Andrey Vlasovskikh
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

r'''A DOT language parser using funcparserlib.

The parser is based on [the DOT grammar][1]. It is pretty complete with a few
not supported things:

* Ports and compass points
* XML identifiers

At the moment, the parser builds only a parse tree, not an abstract syntax tree
(AST) or an API for dealing with DOT.

  [1]: http://www.graphviz.org/doc/info/lang.html
'''

import io
from re import MULTILINE, DOTALL
from collections import namedtuple
from funcparserlib.lexer import make_tokenizer, Token, LexerError
from funcparserlib.parser import (some, a, maybe, many, finished, skip,
                                  forward_decl)
from blockdiag.utils.compat import u


ENCODING = 'utf-8'

Diagram = namedtuple('Diagram', 'id stmts')
Group = namedtuple('Group', 'id stmts')
Node = namedtuple('Node', 'id attrs')
Attr = namedtuple('Attr', 'name value')
Edge = namedtuple('Edge', 'from_nodes edge_type to_nodes attrs')
Extension = namedtuple('Extension', 'type name attrs')
Statements = namedtuple('Statements', 'stmts')


class ParseException(Exception):
    pass


def tokenize(string):
    """str -> Sequence(Token)"""
    # flake8: NOQA
    specs = [                                                                 # NOQA
        ('Comment', (r'/\*(.|[\r\n])*?\*/', MULTILINE)),                      # NOQA
        ('Comment', (r'(//|#).*',)),                                          # NOQA
        ('NL',      (r'[\r\n]+',)),                                           # NOQA
        ('Space',   (r'[ \t\r\n]+',)),                                        # NOQA
        ('Name',    (u('[A-Za-z_0-9\u0080-\uffff]') +                         # NOQA
                     u('[A-Za-z_\\-.0-9\u0080-\uffff]*'),)),                  # NOQA
        ('Op',      (r'[{};,=\[\]]|(<->)|(<-)|(--)|(->)|(>-<)|(-<)|(>-)',)),  # NOQA
        ('Number',  (r'-?(\.[0-9]+)|([0-9]+(\.[0-9]*)?)',)),                  # NOQA
        ('String',  (r'(?P<quote>"|\').*?(?<!\\)(?P=quote)', DOTALL)),        # NOQA
    ]
    useless = ['Comment', 'NL', 'Space']
    t = make_tokenizer(specs)
    return [x for x in t(string) if x.type not in useless]


def create_mapper(fn, default_value=None):
    if default_value is None:
        return lambda args: fn(*args)
    else:
        return lambda args: fn(*args) if args else default_value


def flatten(seq):
    return sum(seq, [])


def oneplus_to_list(first, more):
    return [first] + more


def parse(seq):
    """Sequence(Token) -> object"""
    tokval = lambda x: x.value
    op = lambda s: a(Token('Op', s)) >> tokval
    op_ = lambda s: skip(op(s))
    _id = some(lambda t: t.type in ['Name', 'Number', 'String']) >> tokval
    keyword = lambda s: a(Token('Name', s)) >> tokval

    def make_node_list(node_list, attrs):
        return Statements([Node(node, attrs) for node in node_list])

    def make_edge(first, edge_type, second, followers, attrs):
        edges = [Edge(first, edge_type, second, attrs)]

        from_node = second
        for edge_type, to_node in followers:
            edges.append(Edge(from_node, edge_type, to_node, attrs))
            from_node = to_node

        return Statements(edges)

    #
    # parts of syntax
    #
    node_list = (
        _id +
        many(op_(',') + _id)
        >> create_mapper(oneplus_to_list)
    )
    option_stmt = (
        _id +
        maybe(op_('=') + _id)
        >> create_mapper(Attr)
    )
    option_list = (
        maybe(op_('[') + option_stmt + many(op_(',') + option_stmt) + op_(']'))
        >> create_mapper(oneplus_to_list, default_value=[])
    )

    #  node (node list) statement::
    #     A;
    #     B [attr = value, attr = value];
    #     C, D [attr = value, attr = value];
    #
    node_stmt = (
        node_list + option_list
        >> create_mapper(make_node_list)
    )

    #  edge statement::
    #     A -> B;
    #     A <- B;
    #
    edge_relation = (
        op('->') | op('--') | op('<-') | op('<->') |
        op('>-') | op('-<') | op('>-<')
    )
    edge_stmt = (
        node_list +
        edge_relation +
        node_list +
        many(edge_relation + node_list) +
        option_list
        >> create_mapper(make_edge)
    )

    #  attributes statement::
    #     default_shape = box;
    #     default_fontsize = 16;
    #
    attribute_stmt = (
        _id + op_('=') + _id
        >> create_mapper(Attr)
    )

    #  extension statement (class, plugin)::
    #     class red [color = red];
    #     plugin attributes [name = Name];
    #
    extension_stmt = (
        (keyword('class') | keyword('plugin')) +
        _id +
        option_list
        >> create_mapper(Extension)
    )

    #  group statement::
    #     group {
    #        A;
    #     }
    #
    group_stmt = forward_decl()
    group_inline_stmt = (
        edge_stmt |
        group_stmt |
        attribute_stmt |
        node_stmt
    )
    group_inline_stmt_list = (
        many(group_inline_stmt + skip(maybe(op(';'))))
    )
    group_stmt.define(
        skip(keyword('group')) +
        maybe(_id) +
        op_('{') +
        group_inline_stmt_list +
        op_('}')
        >> create_mapper(Group)
    )

    #
    # diagram statement::
    #     blockdiag {
    #        A;
    #     }
    #
    diagram_id = (
        (keyword('diagram') | keyword('blockdiag')) +
        maybe(_id)
        >> list
    )
    diagram_inline_stmt = (
        extension_stmt |
        group_inline_stmt
    )
    diagram_inline_stmt_list = (
        many(diagram_inline_stmt + skip(maybe(op(';'))))
    )
    diagram = (
        maybe(diagram_id) +
        op_('{') +
        diagram_inline_stmt_list +
        op_('}')
        >> create_mapper(Diagram)
    )
    dotfile = diagram + skip(finished)

    return dotfile.parse(seq)


def sort_tree(tree):
    def weight(node):
        if isinstance(node, (Attr, Extension)):
            return 1
        else:
            return 2

    if hasattr(tree, 'stmts'):
        tree.stmts.sort(key=lambda x: weight(x))
        for stmt in tree.stmts:
            sort_tree(stmt)

    return tree


def parse_string(string):
    try:
        tree = parse(tokenize(string))
        return sort_tree(tree)
    except LexerError as e:
        message = "Got unexpected token at line %d column %d" % e.place
        raise ParseException(message)
    except Exception as e:
        raise ParseException(str(e))


def parse_file(path):
    code = io.open(path, 'r', encoding='utf-8-sig').read()
    return parse_string(code)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division
import pkg_resources
from blockdiag.noderenderer.base import NodeShape  # NOQA: backward compatibility

renderers = {}
searchpath = []


def init_renderers():
    for plugin in pkg_resources.iter_entry_points('blockdiag_noderenderer'):
        module = plugin.load()
        if hasattr(module, 'setup'):
            module.setup(module)


def install_renderer(name, renderer):
    renderers[name] = renderer


def set_default_namespace(path):
    searchpath[:] = []
    for path in path.split(','):
        searchpath.append(path)


def get(shape):
    if not renderers:
        init_renderers()

    for path in searchpath:
        name = "%s.%s" % (path, shape)
        if name in renderers:
            return renderers[name]

    return renderers.get(shape)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division
from blockdiag.utils import images, Box, XY


class NodeShape(object):
    def __init__(self, node, metrics):
        self.node = node
        self.metrics = metrics

        m = self.metrics.cell(self.node)
        self.textalign = 'center'
        self.connectors = [m.top, m.right, m.bottom, m.left]

        if node.icon is None:
            self.iconbox = None
            self.textbox = m.box
        else:
            image_size = images.get_image_size(node.icon)
            if image_size is None:
                iconsize = (0, 0)
            else:
                boundedbox = [metrics.node_width // 2, metrics.node_height]
                iconsize = images.calc_image_size(image_size, boundedbox)

            vmargin = (metrics.node_height - iconsize[1]) // 2
            self.iconbox = Box(m.topleft.x,
                               m.topleft.y + vmargin,
                               m.topleft.x + iconsize[0],
                               m.topleft.y + vmargin + iconsize[1])

            self.textbox = Box(self.iconbox[2], m.top.y,
                               m.bottomright.x, m.bottomright.y)

    def render(self, drawer, _format, **kwargs):
        if self.node.stacked and not kwargs.get('stacked'):
            node = self.node.duplicate()
            node.label = ""
            node.background = ""
            for i in range(2, 0, -1):
                # use original_metrics FORCE
                r = self.metrics.original_metrics.cellsize // 2 * i
                metrics = self.metrics.shift(r, r)

                self.__class__(node, metrics).render(drawer, _format,
                                                     stacked=True, **kwargs)

        if hasattr(self, 'render_vector_shape') and _format == 'SVG':
            self.render_vector_shape(drawer, _format, **kwargs)
        else:
            self.render_shape(drawer, _format, **kwargs)

        self.render_icon(drawer, **kwargs)
        self.render_label(drawer, **kwargs)
        self.render_number_badge(drawer, **kwargs)

    def render_icon(self, drawer, **kwargs):
        if self.node.icon is not None and kwargs.get('shadow') is not True:
            drawer.image(self.iconbox, self.node.icon)

    def render_shape(self, drawer, _, **kwargs):
        pass

    def render_label(self, drawer, **kwargs):
        if not kwargs.get('shadow'):
            font = self.metrics.font_for(self.node)
            drawer.textarea(self.textbox, self.node.label, font,
                            rotate=self.node.rotate,
                            fill=self.node.textcolor, halign=self.textalign,
                            line_spacing=self.metrics.line_spacing,
                            orientation=self.node.label_orientation)

    def render_number_badge(self, drawer, **kwargs):
        if self.node.numbered is not None and kwargs.get('shadow') is None:
            badgeFill = kwargs.get('badgeFill')

            xy = self.metrics.cell(self.node).topleft
            r = self.metrics.cellsize * 3 // 2

            box = Box(xy.x - r, xy.y - r, xy.x + r, xy.y + r)
            font = self.metrics.font_for(self.node)
            drawer.ellipse(box, outline=self.node.linecolor, fill=badgeFill)
            drawer.textarea(box, self.node.numbered, font,
                            rotate=self.node.rotate,
                            fill=self.node.textcolor)

    @property
    def top(self):
        return self.connectors[0]

    @property
    def left(self):
        return self.connectors[3]

    @property
    def right(self):
        point = self.connectors[1]
        if self.node.stacked:
            point = XY(point.x + self.metrics.cellsize, point.y)
        return point

    @property
    def bottom(self):
        point = self.connectors[2]
        if self.node.stacked:
            point = XY(point.x, point.y + self.metrics.cellsize)
        return point

    def shift_shadow(self, value):
        xdiff = self.metrics.shadow_offset.x
        ydiff = self.metrics.shadow_offset.y

        if isinstance(value, XY):
            ret = XY(value.x + xdiff, value.y + ydiff)
        elif isinstance(value, Box):
            ret = Box(value.x1 + xdiff, value.y1 + ydiff,
                      value.x2 + xdiff, value.y2 + ydiff)
        elif isinstance(value, (list, tuple)):
            ret = [self.shift_shadow(x) for x in value]

        return ret



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division
from blockdiag.noderenderer import install_renderer
from blockdiag.noderenderer.base import NodeShape
from blockdiag.utils import Box
from blockdiag.imagedraw.simplesvg import pathdata


class Cloud(NodeShape):
    def __init__(self, node, metrics=None):
        super(Cloud, self).__init__(node, metrics)

        pt = metrics.cell(node).topleft
        rx = (self.node.width or self.metrics.node_width) // 12
        ry = (self.node.height or self.metrics.node_height) // 5
        self.textbox = Box(pt.x + rx * 2, pt.y + ry,
                           pt.x + rx * 11, pt.y + ry * 4)

    def render_shape(self, drawer, _, **kwargs):
        # draw background
        self.render_shape_background(drawer, **kwargs)

        if not kwargs.get('shadow') and self.node.background:
            drawer.image(self.textbox, self.node.background)

    def render_shape_background(self, drawer, **kwargs):
        fill = kwargs.get('fill')

        m = self.metrics.cell(self.node)
        pt = m.topleft
        rx = (self.node.width or self.metrics.node_width) // 12
        ry = (self.node.height or self.metrics.node_height) // 5

        ellipses = [Box(pt.x + rx * 2, pt.y + ry,
                        pt.x + rx * 5, pt.y + ry * 3),
                    Box(pt.x + rx * 4, pt.y,
                        pt.x + rx * 9, pt.y + ry * 2),
                    Box(pt.x + rx * 8, pt.y + ry,
                        pt.x + rx * 11, pt.y + ry * 3),
                    Box(pt.x + rx * 9, pt.y + ry * 2,
                        pt.x + rx * 13, pt.y + ry * 4),
                    Box(pt.x + rx * 8, pt.y + ry * 2,
                        pt.x + rx * 11, pt.y + ry * 5),
                    Box(pt.x + rx * 5, pt.y + ry * 2,
                        pt.x + rx * 8, pt.y + ry * 5),
                    Box(pt.x + rx * 2, pt.y + ry * 2,
                        pt.x + rx * 5, pt.y + ry * 5),
                    Box(pt.x + rx * 0, pt.y + ry * 2,
                        pt.x + rx * 4, pt.y + ry * 4)]

        for e in ellipses:
            if kwargs.get('shadow'):
                e = self.shift_shadow(e)
                if kwargs.get('style') == 'blur':
                    drawer.ellipse(e, fill=fill, outline=fill,
                                   filter='transp-blur')
                else:
                    drawer.ellipse(e, fill=fill, outline=fill)
            else:
                drawer.ellipse(e, fill=self.node.color,
                               outline=self.node.linecolor,
                               style=self.node.style)

        rects = [Box(pt.x + rx * 2, pt.y + ry * 2,
                     pt.x + rx * 11, pt.y + ry * 4),
                 Box(pt.x + rx * 4, pt.y + ry,
                     pt.x + rx * 9, pt.y + ry * 2)]
        for rect in rects:
            if kwargs.get('shadow'):
                rect = self.shift_shadow(rect)
                if kwargs.get('style') == 'blur':
                    drawer.rectangle(rect, fill=fill, outline=fill,
                                     filter='transp-blur')
                else:
                    drawer.rectangle(rect, fill=fill, outline=fill)
            else:
                drawer.rectangle(rect, fill=self.node.color,
                                 outline=self.node.color)

    def render_vector_shape(self, drawer, _, **kwargs):
        fill = kwargs.get('fill')

        # create pathdata
        m = self.metrics.cell(self.node)
        rx = (self.node.width or self.metrics.node_width) // 12
        ry = (self.node.height or self.metrics.node_height) // 5

        pt = m.topleft
        if kwargs.get('shadow'):
            pt = self.shift_shadow(pt)

        path = pathdata(pt.x + rx * 2, pt.y + ry * 2)
        path.ellarc(rx * 2, ry, 0, 0, 1, pt.x + rx * 4, pt.y + ry)
        path.ellarc(rx * 2, ry * 3 // 4, 0, 0, 1, pt.x + rx * 9, pt.y + ry)
        path.ellarc(rx * 2, ry, 0, 0, 1, pt.x + rx * 11, pt.y + ry * 2)
        path.ellarc(rx * 2, ry, 0, 0, 1, pt.x + rx * 11, pt.y + ry * 4)
        path.ellarc(rx * 2, ry * 5 // 2, 0, 0, 1, pt.x + rx * 8, pt.y + ry * 4)
        path.ellarc(rx * 2, ry * 5 // 2, 0, 0, 1, pt.x + rx * 5, pt.y + ry * 4)
        path.ellarc(rx * 2, ry * 5 // 2, 0, 0, 1, pt.x + rx * 2, pt.y + ry * 4)
        path.ellarc(rx * 2, ry, 0, 0, 1, pt.x + rx * 2, pt.y + ry * 2)

        # draw outline
        if kwargs.get('shadow'):
            if kwargs.get('style') == 'blur':
                drawer.path(path, fill=fill, outline=fill,
                            filter='transp-blur')
            else:
                drawer.path(path, fill=fill, outline=fill)
        elif self.node.background:
            drawer.path(path, fill=self.node.color, outline=self.node.color)
            drawer.image(self.textbox, self.node.background)
            drawer.path(path, fill="none", outline=self.node.linecolor,
                        style=self.node.style)
        else:
            drawer.path(path, fill=self.node.color,
                        outline=self.node.linecolor, style=self.node.style)


def setup(self):
    install_renderer('cloud', Cloud)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import sys
from blockdiag.command import BlockdiagApp
from blockdiag.tests.utils import TemporaryDirectory

if sys.version_info < (2, 7):
    import unittest2 as unittest
else:
    import unittest


class TestBlockdiagApp(unittest.TestCase):
    def test_app_cleans_up_images(self):
        testdir = os.path.dirname(__file__)
        diagpath = os.path.join(testdir,
                                'diagrams',
                                'background_url_image.diag')
        urlopen_cache = {}

        def cleanup():
            from blockdiag.utils import images
            urlopen_cache.update(images.urlopen_cache)

        try:
            tmpdir = TemporaryDirectory()
            fd, tmpfile = tmpdir.mkstemp()
            os.close(fd)

            args = ['-T', 'SVG', '-o', tmpfile, diagpath]
            app = BlockdiagApp()
            app.register_cleanup_handler(cleanup)  # to get internal state
            app.run(args)

            self.assertTrue(urlopen_cache)  # check images were cached
            for path in urlopen_cache.values():
                self.assertFalse(os.path.exists(path))  # and removed finally
        finally:
            tmpdir.clean()

    def test_app_cleans_up_plugins(self):
        testdir = os.path.dirname(__file__)
        diagpath = os.path.join(testdir,
                                'diagrams',
                                'plugin_autoclass.diag')
        loaded_plugins = []

        def cleanup():
            from blockdiag import plugins
            loaded_plugins.extend(plugins.loaded_plugins)

        try:
            tmpdir = TemporaryDirectory()
            fd, tmpfile = tmpdir.mkstemp()
            os.close(fd)

            args = ['-T', 'SVG', '-o', tmpfile, diagpath]
            app = BlockdiagApp()
            app.register_cleanup_handler(cleanup)  # to get internal state
            app.run(args)

            from blockdiag import plugins
            self.assertTrue(loaded_plugins)  # check plugins were loaded
            self.assertFalse(plugins.loaded_plugins)  # and unloaded finally
        finally:
            tmpdir.clean()



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from collections import defaultdict
from blockdiag import imagedraw, noderenderer
from blockdiag.metrics import AutoScaler, DiagramMetrics


class DiagramDraw(object):
    shadow_colors = defaultdict(lambda: (0, 0, 0))
    shadow_colors['PNG'] = (64, 64, 64)
    shadow_colors['PDF'] = (144, 144, 144)

    def __init__(self, _format, diagram, filename=None, **kwargs):
        self.format = _format.upper()
        self.diagram = diagram
        self.fill = kwargs.get('fill', (0, 0, 0))
        self.badgeFill = kwargs.get('badgeFill', 'pink')
        self.filename = filename
        self.shadow = self.shadow_colors[self.format.upper()]

        if self.format == 'PNG' and kwargs.get('antialias'):
            self.scale_ratio = 2
        else:
            self.scale_ratio = 1

        self.drawer = imagedraw.create(self.format, self.filename,
                                       filters=['linejump'],
                                       scale_ratio=self.scale_ratio,
                                       **kwargs)

        self.metrics = self.create_metrics(kwargs.get('basediagram', diagram),
                                           drawer=self.drawer,
                                           fontmap=kwargs.get('fontmap'))
        if self.scale_ratio == 2:
            self.metrics = AutoScaler(self.metrics,
                                      scale_ratio=self.scale_ratio)

        self.drawer.set_canvas_size(self.pagesize())
        self.drawer.set_options(jump_radius=self.metrics.cellsize / 2)

    def create_metrics(self, *args, **kwargs):
        return DiagramMetrics(*args, **kwargs)

    @property
    def nodes(self):
        for node in self.diagram.traverse_nodes():
            if node.drawable:
                yield node

    @property
    def groups(self):
        for group in self.diagram.traverse_groups(preorder=True):
            if not group.drawable:
                yield group

    @property
    def edges(self):
        edges = self.diagram.traverse_edges(preorder=True)
        for edge in (e for e in edges if e.style != 'none'):
            yield edge

    def pagesize(self, scaled=False):
        if scaled:
            metrics = self.metrics
        else:
            metrics = self.metrics.original_metrics

        width = self.diagram.colwidth
        height = self.diagram.colheight
        return metrics.pagesize(width, height)

    def draw(self, **kwargs):
        # switch metrics object during draw backgrounds
        temp, self.metrics = self.metrics, self.metrics.original_metrics
        self._draw_background()
        self.metrics = temp

        if self.scale_ratio > 1:
            pagesize = self.pagesize(scaled=True)
            self.drawer.resizeCanvas(pagesize)

        self._draw_elements(**kwargs)

    def _draw_background(self):
        # Draw node groups.
        for group in self.groups:
            if group.shape == 'box':
                box = self.metrics.group(group).marginbox
                if group.href and self.format == 'SVG':
                    drawer = self.drawer.anchor(group.href)
                else:
                    drawer = self.drawer

                drawer.rectangle(box, fill=group.color, filter='blur')

        # Drop node shadows.
        for node in self.nodes:
            if node.color != 'none' and self.diagram.shadow_style != 'none':
                r = noderenderer.get(node.shape)

                shape = r(node, self.metrics)
                if node.href and self.format == 'SVG':
                    drawer = self.drawer.anchor(node.href)
                else:
                    drawer = self.drawer

                shape.render(drawer, self.format,
                             fill=self.shadow, shadow=True,
                             style=self.diagram.shadow_style)

    def _draw_elements(self, **kwargs):
        for node in self.nodes:
            self.node(node, **kwargs)

        for edge in self.edges:
            self.edge(edge)

        for edge in self.edges:
            self.edge_label(edge)

        for group in self.groups:
            if group.shape == 'line':
                box = self.metrics.group(group).marginbox
                self.drawer.rectangle(box, fill='none', outline=group.color,
                                      style=group.style, thick=group.thick)

        for node in self.groups:
            self.group_label(node, **kwargs)

    def node(self, node, **kwargs):
        r = noderenderer.get(node.shape)
        shape = r(node, self.metrics)
        if node.href and self.format == 'SVG':
            drawer = self.drawer.anchor(node.href)
        else:
            drawer = self.drawer

        shape.render(drawer, self.format, fill=self.fill,
                     badgeFill=self.badgeFill)

    def group_label(self, group):
        m = self.metrics.group(group)
        font = self.metrics.font_for(group)

        if group.label and not group.separated:
            self.drawer.textarea(m.grouplabelbox, group.label, font=font,
                                 fill=group.textcolor)
        elif group.label:
            self.drawer.textarea(m.corebox, group.label, font=font,
                                 fill=group.textcolor)

    def edge(self, edge):
        metrics = self.metrics.edge(edge)

        for line in metrics.shaft.polylines:
            self.drawer.line(line, fill=edge.color, thick=edge.thick,
                             style=edge.style, jump=True)

        for head in metrics.heads:
            if edge.hstyle in ('generalization', 'aggregation'):
                self.drawer.polygon(head, outline=edge.color, fill='white')
            else:
                self.drawer.polygon(head, outline=edge.color, fill=edge.color)

    def edge_label(self, edge):
        if edge.label:
            metrics = self.metrics.edge(edge)

            font = self.metrics.font_for(edge)
            self.drawer.textarea(metrics.labelbox, edge.label, font=font,
                                 fill=edge.textcolor, outline=self.fill)

    def save(self, size=None):
        return self.drawer.save(self.filename, size, self.format)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pkg_resources import iter_entry_points
from blockdiag.utils.logging import warning

loaded_plugins = []
node_handlers = []
general_handlers = {}


def load(plugins, diagram, **kwargs):
    for name in plugins:
        if name in loaded_plugins:
            warning('plugin "%s" is already loaded. ignored.', name)
            return

        for ep in iter_entry_points('blockdiag_plugins', name):
            module = ep.load()
            loaded_plugins.append(name)
            if hasattr(module, 'setup'):
                module.setup(module, diagram, **kwargs)
            break
        else:
            msg = "unknown plugin: %s" % name
            raise AttributeError(msg)


def install_general_handler(name, handler):
    if name not in general_handlers:
        general_handlers[name] = []

    general_handlers[name].append(handler)


def fire_general_event(name, *args):
    handlers = general_handlers.get(name, [])
    return all(handler(*args) for handler in handlers)


def install_node_handler(handler):
    if handler not in node_handlers:
        node_handlers.append(handler)


def fire_node_event(node, name, *args):
    return all(handler.fire(name, node, *args) for handler in node_handlers)


class NodeHandler(object):
    def __init__(self, diagram, **kwargs):
        self.diagram = diagram
        self.config = kwargs.get('config')

    def fire(self, name, *args):
        return getattr(self, "on_" + name)(*args)

    def on_created(self, node):
        return True

    def on_attr_changing(self, node, attr):
        return True

    def on_attr_changed(self, node, attr):
        return True

    def on_build_finished(self, node):
        return True


def cleanup():
    fire_general_event('cleanup')

    for name in list(general_handlers.keys()):
        del general_handlers[name]

    for handler in node_handlers[:]:
        node_handlers.remove(handler)

    for plugin in loaded_plugins[:]:
        loaded_plugins.remove(plugin)


def setup(app):
    app.register_cleanup_handler(cleanup)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from blockdiag.tests.utils import BuilderTestCase
from blockdiag.parser import ParseException


class TestBuilderError(BuilderTestCase):
    def test_unknown_diagram_default_shape_diagram(self):
        filename = 'errors/unknown_diagram_default_shape.diag'
        with self.assertRaises(AttributeError):
            self.build(filename)

    def test_unknown_diagram_edge_layout_diagram(self):
        filename = 'errors/unknown_diagram_edge_layout.diag'
        with self.assertRaises(AttributeError):
            self.build(filename)

    def test_unknown_diagram_orientation_diagram(self):
        filename = 'errors/unknown_diagram_orientation.diag'
        with self.assertRaises(AttributeError):
            self.build(filename)

    def test_unknown_node_shape_diagram(self):
        filename = 'errors/unknown_node_shape.diag'
        with self.assertRaises(AttributeError):
            self.build(filename)

    def test_unknown_node_attribute_diagram(self):
        filename = 'errors/unknown_node_attribute.diag'
        with self.assertRaises(AttributeError):
            self.build(filename)

    def test_unknown_node_style_diagram(self):
        filename = 'errors/unknown_node_style.diag'
        with self.assertRaises(AttributeError):
            self.build(filename)

    def test_unknown_node_class_diagram(self):
        filename = 'errors/unknown_node_class.diag'
        with self.assertRaises(AttributeError):
            self.build(filename)

    def test_unknown_edge_dir_diagram(self):
        filename = 'errors/unknown_edge_dir.diag'
        with self.assertRaises(AttributeError):
            self.build(filename)

    def test_unknown_edge_style_diagram(self):
        filename = 'errors/unknown_edge_style.diag'
        with self.assertRaises(AttributeError):
            self.build(filename)

    def test_unknown_edge_hstyle_diagram(self):
        filename = 'errors/unknown_edge_hstyle.diag'
        with self.assertRaises(AttributeError):
            self.build(filename)

    def test_unknown_edge_class_diagram(self):
        filename = 'errors/unknown_edge_class.diag'
        with self.assertRaises(AttributeError):
            self.build(filename)

    def test_unknown_group_shape_diagram(self):
        filename = 'errors/unknown_group_shape.diag'
        with self.assertRaises(AttributeError):
            self.build(filename)

    def test_unknown_group_class_diagram(self):
        filename = 'errors/unknown_group_class.diag'
        with self.assertRaises(AttributeError):
            self.build(filename)

    def test_unknown_group_orientation_diagram(self):
        filename = 'errors/unknown_group_orientation.diag'
        with self.assertRaises(AttributeError):
            self.build(filename)

    def test_belongs_to_two_groups_diagram(self):
        filename = 'errors/belongs_to_two_groups.diag'
        with self.assertRaises(RuntimeError):
            self.build(filename)

    def test_unknown_plugin_diagram(self):
        filename = 'errors/unknown_plugin.diag'
        with self.assertRaises(AttributeError):
            self.build(filename)

    def test_node_follows_group_diagram(self):
        filename = 'errors/node_follows_group.diag'
        with self.assertRaises(ParseException):
            self.build(filename)

    def test_group_follows_node_diagram(self):
        filename = 'errors/group_follows_node.diag'
        with self.assertRaises(ParseException):
            self.build(filename)

    def test_unknown_diagram_type(self):
        filename = 'errors/unknown_diagram_type.diag'
        with self.assertRaises(ParseException):
            self.build(filename)

    def test_lexer_error_diagram(self):
        filename = 'errors/lexer_error.diag'
        with self.assertRaises(ParseException):
            self.build(filename)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import sys
from docutils import nodes
from docutils.core import publish_doctree
from docutils.parsers.rst import directives as docutils
from blockdiag.utils.compat import u
from blockdiag.utils.rst import directives
from blockdiag.tests.utils import capture_stderr, with_pil, TemporaryDirectory


if sys.version_info < (2, 7):
    import unittest2 as unittest
else:
    import unittest


class TestRstDirectives(unittest.TestCase):
    def setUp(self):
        self._tmpdir = TemporaryDirectory()

    def tearDown(self):
        if 'blockdiag' in docutils._directives:
            del docutils._directives['blockdiag']

        self._tmpdir.clean()

    @property
    def tmpdir(self):
        return self._tmpdir.name

    def test_setup(self):
        directives.setup()
        options = directives.directive_options

        self.assertIn('blockdiag', docutils._directives)
        self.assertEqual(directives.BlockdiagDirective,
                         docutils._directives['blockdiag'])
        self.assertEqual('PNG', options['format'])
        self.assertEqual(False, options['antialias'])
        self.assertEqual(None, options['fontpath'])
        self.assertEqual(False, options['nodoctype'])
        self.assertEqual(False, options['noviewbox'])
        self.assertEqual(False, options['inline_svg'])

    def test_setup_with_args(self):
        directives.setup(format='SVG', antialias=True, fontpath='/dev/null',
                         nodoctype=True, noviewbox=True, inline_svg=True)
        options = directives.directive_options

        self.assertIn('blockdiag', docutils._directives)
        self.assertEqual(directives.BlockdiagDirective,
                         docutils._directives['blockdiag'])
        self.assertEqual('SVG', options['format'])
        self.assertEqual(True, options['antialias'])
        self.assertEqual('/dev/null', options['fontpath'])
        self.assertEqual(True, options['nodoctype'])
        self.assertEqual(True, options['noviewbox'])
        self.assertEqual(True, options['inline_svg'])

    @capture_stderr
    def test_cleanup(self):
        directives.setup(format='SVG', outputdir=self.tmpdir, noviewbox=False)
        text = (".. blockdiag::\n"
                "\n"
                "   plugin autoclass\n"
                "   A -> B")
        publish_doctree(text)

        from blockdiag import plugins
        self.assertEqual([], plugins.loaded_plugins)

    def test_setup_fontpath1(self):
        with self.assertRaises(RuntimeError):
            directives.setup(format='SVG', fontpath=['dummy.ttf'],
                             outputdir=self.tmpdir)
            text = (".. blockdiag::\n"
                    "\n"
                    "   A -> B")
            publish_doctree(text)

    def test_setup_fontpath2(self):
        with self.assertRaises(RuntimeError):
            directives.setup(format='SVG', fontpath='dummy.ttf',
                             outputdir=self.tmpdir)
            text = (".. blockdiag::\n"
                    "\n"
                    "   A -> B")
            publish_doctree(text)

    def test_setup_nodoctype_is_true(self):
        directives.setup(format='SVG', outputdir=self.tmpdir, nodoctype=True)
        text = (".. blockdiag::\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.image, type(doctree[-1]))
        svg = open(doctree[0]['uri']).read()
        self.assertEqual("<svg viewBox=\"0 0 ", svg[:18])

    def test_setup_nodoctype_is_false(self):
        directives.setup(format='SVG', outputdir=self.tmpdir, nodoctype=False)
        text = (".. blockdiag::\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))
        svg = open(doctree[0]['uri']).read()
        self.assertEqual("<?xml version='1.0' encoding='UTF-8'?>\n"
                         "<!DOCTYPE ", svg[:49])

    def test_setup_noviewbox_is_true(self):
        directives.setup(format='SVG', outputdir=self.tmpdir, noviewbox=True)
        text = (".. blockdiag::\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))
        svg = open(doctree[0]['uri']).read()
        self.assertRegexpMatches(svg, '<svg height="\d+" width="\d+" ')

    def test_setup_noviewbox_is_false(self):
        directives.setup(format='SVG', outputdir=self.tmpdir, noviewbox=False)
        text = (".. blockdiag::\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))
        svg = open(doctree[0]['uri']).read()
        self.assertRegexpMatches(svg, '<svg viewBox="0 0 \d+ \d+" ')

    def test_setup_inline_svg_is_true(self):
        directives.setup(format='SVG', outputdir=self.tmpdir, inline_svg=True)
        text = (".. blockdiag::\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.raw, type(doctree[0]))
        self.assertEqual('html', doctree[0]['format'])
        self.assertEqual(nodes.Text, type(doctree[0][0]))
        self.assertEqual("<?xml version='1.0' encoding='UTF-8'?>\n"
                         "<!DOCTYPE ", doctree[0][0][:49])
        self.assertEqual(0, len(os.listdir(self.tmpdir)))

    def test_setup_inline_svg_is_false(self):
        directives.setup(format='SVG', outputdir=self.tmpdir, inline_svg=False)
        text = (".. blockdiag::\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))
        self.assertEqual(1, len(os.listdir(self.tmpdir)))

    @with_pil
    def test_setup_inline_svg_is_true_but_format_isnt_svg(self):
        directives.setup(format='PNG', outputdir=self.tmpdir, inline_svg=True)
        text = (".. blockdiag::\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))

    def test_setup_inline_svg_is_true_with_multibytes(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = u(".. blockdiag::\n"
                 "\n"
                 "   あ -> い")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))

    def test_setup_inline_svg_is_true_and_width_option1(self):
        directives.setup(format='SVG', outputdir=self.tmpdir,
                         nodoctype=True, noviewbox=True, inline_svg=True)
        text = (".. blockdiag::\n"
                "   :width: 100\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.raw, type(doctree[0]))
        self.assertEqual(nodes.Text, type(doctree[0][0]))
        self.assertRegexpMatches(doctree[0][0],
                                 '<svg height="\d+" width="100" ')

    def test_setup_inline_svg_is_true_and_width_option2(self):
        directives.setup(format='SVG', outputdir=self.tmpdir,
                         nodoctype=True, noviewbox=True, inline_svg=True)
        text = (".. blockdiag::\n"
                "   :width: 10000\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.raw, type(doctree[0]))
        self.assertEqual(nodes.Text, type(doctree[0][0]))
        self.assertRegexpMatches(doctree[0][0],
                                 '<svg height="\d+" width="10000" ')

    def test_setup_inline_svg_is_true_and_height_option1(self):
        directives.setup(format='SVG', outputdir=self.tmpdir,
                         nodoctype=True, noviewbox=True, inline_svg=True)
        text = (".. blockdiag::\n"
                "   :height: 100\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.raw, type(doctree[0]))
        self.assertEqual(nodes.Text, type(doctree[0][0]))
        self.assertRegexpMatches(doctree[0][0],
                                 '<svg height="100" width="\d+" ')

    def test_setup_inline_svg_is_true_and_height_option2(self):
        directives.setup(format='SVG', outputdir=self.tmpdir,
                         nodoctype=True, noviewbox=True, inline_svg=True)
        text = (".. blockdiag::\n"
                "   :height: 10000\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.raw, type(doctree[0]))
        self.assertEqual(nodes.Text, type(doctree[0][0]))
        self.assertRegexpMatches(doctree[0][0],
                                 '<svg height="10000" width="\d+" ')

    def test_setup_inline_svg_is_true_and_width_and_height_option(self):
        directives.setup(format='SVG', outputdir=self.tmpdir,
                         nodoctype=True, noviewbox=True, inline_svg=True)
        text = (".. blockdiag::\n"
                "   :width: 200\n"
                "   :height: 100\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.raw, type(doctree[0]))
        self.assertEqual(nodes.Text, type(doctree[0][0]))
        self.assertRegexpMatches(doctree[0][0],
                                 '<svg height="100" width="200" ')

    def test_call_with_brace(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "\n"
                "   {"
                "     A -> B"
                "   }")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))
        self.assertEqual(0, doctree[0]['uri'].index(self.tmpdir))

    def test_call_without_braces(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))
        self.assertEqual(0, doctree[0]['uri'].index(self.tmpdir))

    def test_alt_option(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :alt: hello world\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))
        self.assertEqual('hello world', doctree[0]['alt'])
        self.assertEqual(0, doctree[0]['uri'].index(self.tmpdir))

    def test_align_option1(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :align: left\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))
        self.assertEqual('left', doctree[0]['align'])
        self.assertEqual(0, doctree[0]['uri'].index(self.tmpdir))

    def test_align_option2(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :align: center\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))
        self.assertEqual('center', doctree[0]['align'])
        self.assertEqual(0, doctree[0]['uri'].index(self.tmpdir))

    def test_align_option3(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :align: right\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))
        self.assertEqual('right', doctree[0]['align'])
        self.assertEqual(0, doctree[0]['uri'].index(self.tmpdir))

    @capture_stderr
    def test_align_option4(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :align: unknown\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.system_message, type(doctree[0]))

        # clear stderr outputs (ignore ERROR)
        from io import StringIO
        sys.stderr = StringIO()

    def test_caption_option(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :caption: hello world\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.figure, type(doctree[0]))
        self.assertEqual(2, len(doctree[0]))
        self.assertEqual(nodes.image, type(doctree[0][0]))
        self.assertEqual(nodes.caption, type(doctree[0][1]))
        self.assertEqual(1, len(doctree[0][1]))
        self.assertEqual(nodes.Text, type(doctree[0][1][0]))
        self.assertEqual('hello world', doctree[0][1][0])

    def test_caption_option2(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :caption: **hello** *world*\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.figure, type(doctree[0]))
        self.assertEqual(2, len(doctree[0]))
        self.assertEqual(nodes.image, type(doctree[0][0]))
        self.assertEqual(nodes.caption, type(doctree[0][1]))
        self.assertEqual(3, len(doctree[0][1]))
        self.assertEqual(nodes.strong, type(doctree[0][1][0]))
        self.assertEqual('hello', doctree[0][1][0][0])
        self.assertEqual(nodes.Text, type(doctree[0][1][1]))
        self.assertEqual(' ', doctree[0][1][1][0])
        self.assertEqual(nodes.emphasis, type(doctree[0][1][2]))
        self.assertEqual('world', doctree[0][1][2][0])

    def test_caption_option_and_align_option(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :align: left\n"
                "   :caption: hello world\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.figure, type(doctree[0]))
        self.assertEqual('left', doctree[0]['align'])
        self.assertEqual(2, len(doctree[0]))
        self.assertEqual(nodes.image, type(doctree[0][0]))
        self.assertNotIn('align', doctree[0][0])
        self.assertEqual(nodes.caption, type(doctree[0][1]))
        self.assertEqual(1, len(doctree[0][1]))
        self.assertEqual(nodes.Text, type(doctree[0][1][0]))
        self.assertEqual('hello world', doctree[0][1][0])

    @capture_stderr
    def test_maxwidth_option(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :maxwidth: 100\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(2, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))
        self.assertEqual('100', doctree[0]['width'])
        self.assertEqual(0, doctree[0]['uri'].index(self.tmpdir))
        self.assertEqual(nodes.system_message, type(doctree[1]))

    def test_width_option(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :width: 100\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))
        self.assertEqual('100', doctree[0]['width'])
        self.assertEqual(0, doctree[0]['uri'].index(self.tmpdir))

    def test_height_option(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :height: 100\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))
        self.assertEqual('100', doctree[0]['height'])
        self.assertEqual(0, doctree[0]['uri'].index(self.tmpdir))

    def test_scale_option(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :scale: 50%\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))
        self.assertEqual(50, doctree[0]['scale'])
        self.assertEqual(0, doctree[0]['uri'].index(self.tmpdir))

    def test_name_option(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :name: foo%\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))
        self.assertEqual(['foo%'], doctree[0]['names'])
        self.assertEqual(0, doctree[0]['uri'].index(self.tmpdir))

    def test_class_option(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :class: bar%\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))
        self.assertEqual(['bar'], doctree[0]['classes'])
        self.assertEqual(0, doctree[0]['uri'].index(self.tmpdir))

    def test_figwidth_option1(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :caption: hello world\n"
                "   :figwidth: 100\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.figure, type(doctree[0]))
        self.assertEqual('100px', doctree[0]['width'])

    def test_figwidth_option2(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :caption: hello world\n"
                "   :figwidth: image\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.figure, type(doctree[0]))
        self.assertEqual('448px', doctree[0]['width'])

    def test_figclass_option(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :caption: hello world\n"
                "   :figclass: baz\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.figure, type(doctree[0]))
        self.assertEqual(['baz'], doctree[0]['classes'])

    def test_desctable_option(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :desctable:\n"
                "\n"
                "   A [description = foo]"
                "   B [description = bar]"
                "   group { A }")
        doctree = publish_doctree(text)
        self.assertEqual(2, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))
        self.assertEqual(nodes.table, type(doctree[1]))

        # tgroup
        self.assertEqual(4, len(doctree[1][0]))
        self.assertEqual(nodes.colspec, type(doctree[1][0][0]))
        self.assertEqual(nodes.colspec, type(doctree[1][0][1]))
        self.assertEqual(nodes.thead, type(doctree[1][0][2]))
        self.assertEqual(nodes.tbody, type(doctree[1][0][3]))

        # colspec
        self.assertEqual(50, doctree[1][0][0]['colwidth'])
        self.assertEqual(50, doctree[1][0][1]['colwidth'])

        # thead
        thead = doctree[1][0][2]
        self.assertEqual(2, len(thead[0]))
        self.assertEqual('Name', thead[0][0][0][0])
        self.assertEqual('Description', thead[0][1][0][0])

        # tbody
        tbody = doctree[1][0][3]
        self.assertEqual(2, len(tbody))
        self.assertEqual('A', tbody[0][0][0][0])
        self.assertEqual('foo', tbody[0][1][0][0])
        self.assertEqual('B', tbody[1][0][0][0])
        self.assertEqual('bar', tbody[1][1][0][0])

    def test_desctable_option_without_description(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :desctable:\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))

    def test_desctable_option_using_node_group(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :desctable:\n"
                "\n"
                "   A [description = foo]"
                "   B [description = bar]"
                "   group { A }")
        doctree = publish_doctree(text)
        self.assertEqual(2, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))
        self.assertEqual(nodes.table, type(doctree[1]))

        self.assertEqual(1, len(doctree[1]))
        self.assertEqual(nodes.tgroup, type(doctree[1][0]))

        # tgroup
        self.assertEqual(4, len(doctree[1][0]))
        self.assertEqual(nodes.colspec, type(doctree[1][0][0]))
        self.assertEqual(nodes.colspec, type(doctree[1][0][1]))
        self.assertEqual(nodes.thead, type(doctree[1][0][2]))
        self.assertEqual(nodes.tbody, type(doctree[1][0][3]))

        # colspec
        self.assertEqual(0, len(doctree[1][0][0]))
        self.assertEqual(50, doctree[1][0][0]['colwidth'])

        self.assertEqual(0, len(doctree[1][0][1]))
        self.assertEqual(50, doctree[1][0][1]['colwidth'])

        # thead
        thead = doctree[1][0][2]
        self.assertEqual(1, len(thead))
        self.assertEqual(2, len(thead[0]))

        self.assertEqual(1, len(thead[0][0]))
        self.assertEqual(1, len(thead[0][0][0]))
        self.assertEqual('Name', thead[0][0][0][0])

        self.assertEqual(1, len(thead[0][1]))
        self.assertEqual(1, len(thead[0][1][0]))
        self.assertEqual('Description', thead[0][1][0][0])

        # tbody
        tbody = doctree[1][0][3]
        self.assertEqual(2, len(tbody))

        self.assertEqual(2, len(tbody[0]))
        self.assertEqual(1, len(tbody[0][0]))
        self.assertEqual(1, len(tbody[0][0][0]))
        self.assertEqual('A', tbody[0][0][0][0])
        self.assertEqual(1, len(tbody[0][1]))
        self.assertEqual('foo', tbody[0][1][0][0])

        self.assertEqual(2, len(tbody[1]))
        self.assertEqual(1, len(tbody[1][0]))
        self.assertEqual(1, len(tbody[1][0][0]))
        self.assertEqual('B', tbody[1][0][0][0])
        self.assertEqual(1, len(tbody[1][1]))
        self.assertEqual('bar', tbody[1][1][0][0])

    def test_desctable_option_with_rest_markups(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :desctable:\n"
                "\n"
                "   A [description = \"foo *bar* **baz**\"]"
                "   B [description = \"**foo** *bar* baz\"]")
        doctree = publish_doctree(text)
        self.assertEqual(2, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))
        self.assertEqual(nodes.table, type(doctree[1]))

        # tgroup
        self.assertEqual(4, len(doctree[1][0]))
        self.assertEqual(nodes.colspec, type(doctree[1][0][0]))
        self.assertEqual(nodes.colspec, type(doctree[1][0][1]))
        self.assertEqual(nodes.thead, type(doctree[1][0][2]))
        self.assertEqual(nodes.tbody, type(doctree[1][0][3]))

        # colspec
        self.assertEqual(50, doctree[1][0][0]['colwidth'])
        self.assertEqual(50, doctree[1][0][1]['colwidth'])

        # thead
        thead = doctree[1][0][2]
        self.assertEqual(2, len(thead[0]))
        self.assertEqual('Name', thead[0][0][0][0])
        self.assertEqual('Description', thead[0][1][0][0])

        # tbody
        tbody = doctree[1][0][3]
        self.assertEqual(2, len(tbody))
        self.assertEqual('A', tbody[0][0][0][0])
        self.assertEqual(4, len(tbody[0][1][0]))
        self.assertEqual(nodes.Text, type(tbody[0][1][0][0]))
        self.assertEqual('foo ', str(tbody[0][1][0][0]))
        self.assertEqual(nodes.emphasis, type(tbody[0][1][0][1]))
        self.assertEqual(nodes.Text, type(tbody[0][1][0][1][0]))
        self.assertEqual('bar', tbody[0][1][0][1][0])
        self.assertEqual(nodes.Text, type(tbody[0][1][0][2]))
        self.assertEqual(' ', str(tbody[0][1][0][2]))
        self.assertEqual(nodes.strong, type(tbody[0][1][0][3]))
        self.assertEqual(nodes.Text, type(tbody[0][1][0][3][0]))
        self.assertEqual('baz', str(tbody[0][1][0][3][0]))

        self.assertEqual('B', tbody[1][0][0][0])
        self.assertEqual(4, len(tbody[1][1][0]))
        self.assertEqual(nodes.strong, type(tbody[1][1][0][0]))
        self.assertEqual(nodes.Text, type(tbody[1][1][0][0][0]))
        self.assertEqual('foo', str(tbody[1][1][0][0][0]))
        self.assertEqual(nodes.Text, type(tbody[1][1][0][1]))
        self.assertEqual(' ', str(tbody[1][1][0][1]))
        self.assertEqual(nodes.emphasis, type(tbody[1][1][0][2]))
        self.assertEqual(nodes.Text, type(tbody[1][1][0][2][0]))
        self.assertEqual('bar', str(tbody[1][1][0][2][0]))
        self.assertEqual(nodes.Text, type(tbody[1][1][0][3]))
        self.assertEqual(' baz', str(tbody[1][1][0][3]))

    def test_desctable_option_with_numbered(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :desctable:\n"
                "\n"
                "   A [numbered = 2]"
                "   B [numbered = 1]")
        doctree = publish_doctree(text)
        self.assertEqual(2, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))
        self.assertEqual(nodes.table, type(doctree[1]))

        # tgroup
        self.assertEqual(4, len(doctree[1][0]))
        self.assertEqual(nodes.colspec, type(doctree[1][0][0]))
        self.assertEqual(nodes.colspec, type(doctree[1][0][1]))
        self.assertEqual(nodes.thead, type(doctree[1][0][2]))
        self.assertEqual(nodes.tbody, type(doctree[1][0][3]))

        # colspec
        self.assertEqual(25, doctree[1][0][0]['colwidth'])
        self.assertEqual(50, doctree[1][0][1]['colwidth'])

        # thead
        thead = doctree[1][0][2]
        self.assertEqual(2, len(thead[0]))
        self.assertEqual('No', thead[0][0][0][0])
        self.assertEqual('Name', thead[0][1][0][0])

        # tbody
        tbody = doctree[1][0][3]
        self.assertEqual(2, len(tbody))
        self.assertEqual('1', tbody[0][0][0][0])
        self.assertEqual('B', tbody[0][1][0][0])
        self.assertEqual('2', tbody[1][0][0][0])
        self.assertEqual('A', tbody[1][1][0][0])

    def test_desctable_option_with_numbered_and_description(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :desctable:\n"
                "\n"
                "   A [description = foo, numbered = 2]"
                "   B [description = bar, numbered = 1]")
        doctree = publish_doctree(text)
        self.assertEqual(2, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))
        self.assertEqual(nodes.table, type(doctree[1]))

        # tgroup
        self.assertEqual(5, len(doctree[1][0]))
        self.assertEqual(nodes.colspec, type(doctree[1][0][0]))
        self.assertEqual(nodes.colspec, type(doctree[1][0][1]))
        self.assertEqual(nodes.colspec, type(doctree[1][0][2]))
        self.assertEqual(nodes.thead, type(doctree[1][0][3]))
        self.assertEqual(nodes.tbody, type(doctree[1][0][4]))

        # colspec
        self.assertEqual(25, doctree[1][0][0]['colwidth'])
        self.assertEqual(50, doctree[1][0][1]['colwidth'])
        self.assertEqual(50, doctree[1][0][2]['colwidth'])

        # thead
        thead = doctree[1][0][3]
        self.assertEqual(3, len(thead[0]))
        self.assertEqual('No', thead[0][0][0][0])
        self.assertEqual('Name', thead[0][1][0][0])
        self.assertEqual('Description', thead[0][2][0][0])

        # tbody
        tbody = doctree[1][0][4]
        self.assertEqual(2, len(tbody))
        self.assertEqual('1', tbody[0][0][0][0])
        self.assertEqual('B', tbody[0][1][0][0])
        self.assertEqual(1, len(tbody[0][2]))
        self.assertEqual('bar', tbody[0][2][0][0])
        self.assertEqual('2', tbody[1][0][0][0])
        self.assertEqual('A', tbody[1][1][0][0])
        self.assertEqual('foo', tbody[1][2][0][0])

    def test_desctable_option_for_edges(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :desctable:\n"
                "\n"
                "   A -> B [description = \"foo\"]"
                "   C -> D [description = \"bar\"]"
                "   C [label = \"label_C\"]"
                "   D [label = \"label_D\"]")
        doctree = publish_doctree(text)
        self.assertEqual(2, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))
        self.assertEqual(nodes.table, type(doctree[1]))

        # tgroup
        self.assertEqual(4, len(doctree[1][0]))
        self.assertEqual(nodes.colspec, type(doctree[1][0][0]))
        self.assertEqual(nodes.colspec, type(doctree[1][0][1]))
        self.assertEqual(nodes.thead, type(doctree[1][0][2]))
        self.assertEqual(nodes.tbody, type(doctree[1][0][3]))

        # colspec
        self.assertEqual(25, doctree[1][0][0]['colwidth'])
        self.assertEqual(50, doctree[1][0][1]['colwidth'])

        # thead
        thead = doctree[1][0][2]
        self.assertEqual(2, len(thead[0]))
        self.assertEqual('Name', thead[0][0][0][0])
        self.assertEqual('Description', thead[0][1][0][0])

        # tbody
        tbody = doctree[1][0][3]
        self.assertEqual(2, len(tbody))
        self.assertEqual('A -> B', tbody[0][0][0][0])
        self.assertEqual(1, len(tbody[0][1][0]))
        self.assertEqual(nodes.Text, type(tbody[0][1][0][0]))
        self.assertEqual('foo', str(tbody[0][1][0][0]))
        self.assertEqual('label_C -> label_D', tbody[1][0][0][0])
        self.assertEqual(1, len(tbody[1][1][0]))
        self.assertEqual(nodes.Text, type(tbody[1][1][0][0]))
        self.assertEqual('bar', str(tbody[1][1][0][0]))

    def test_desctable_option_for_nodes_and_edges(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "   :desctable:\n"
                "\n"
                "   A -> B [description = \"foo\"]"
                "   C -> D [description = \"bar\"]"
                "   C [label = \"label_C\", description = foo]"
                "   D [label = \"label_D\"]")
        doctree = publish_doctree(text)
        self.assertEqual(3, len(doctree))
        self.assertEqual(nodes.image, type(doctree[0]))
        self.assertEqual(nodes.table, type(doctree[1]))
        self.assertEqual(nodes.table, type(doctree[2]))

        # tgroup
        self.assertEqual(4, len(doctree[2][0]))
        self.assertEqual(nodes.colspec, type(doctree[2][0][0]))
        self.assertEqual(nodes.colspec, type(doctree[2][0][1]))
        self.assertEqual(nodes.thead, type(doctree[2][0][2]))
        self.assertEqual(nodes.tbody, type(doctree[2][0][3]))

        # colspec
        self.assertEqual(25, doctree[2][0][0]['colwidth'])
        self.assertEqual(50, doctree[2][0][1]['colwidth'])

        # thead
        thead = doctree[2][0][2]
        self.assertEqual(2, len(thead[0]))
        self.assertEqual('Name', thead[0][0][0][0])
        self.assertEqual('Description', thead[0][1][0][0])

        # tbody
        tbody = doctree[2][0][3]
        self.assertEqual(2, len(tbody))
        self.assertEqual('A -> B', tbody[0][0][0][0])
        self.assertEqual(1, len(tbody[0][1][0]))
        self.assertEqual(nodes.Text, type(tbody[0][1][0][0]))
        self.assertEqual('foo', str(tbody[0][1][0][0]))
        self.assertEqual('label_C -> label_D', tbody[1][0][0][0])
        self.assertEqual(1, len(tbody[1][1][0]))
        self.assertEqual(nodes.Text, type(tbody[1][1][0][0]))
        self.assertEqual('bar', str(tbody[1][1][0][0]))

    @capture_stderr
    def test_broken_diagram(self):
        directives.setup(format='SVG', outputdir=self.tmpdir)
        text = (".. blockdiag::\n"
                "\n"
                "     A ->")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.system_message, type(doctree[0]))



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import io
from hashlib import sha1
from functools import wraps
from collections import namedtuple
from docutils import nodes
from docutils.parsers import rst
from docutils.parsers.rst.roles import set_classes
from docutils.statemachine import ViewList

from blockdiag.utils.bootstrap import create_fontmap, Application
from blockdiag.utils.compat import string_types
from blockdiag.utils.rst.nodes import blockdiag as blockdiag_node

directive_options_default = dict(format='PNG',
                                 antialias=False,
                                 fontpath=None,
                                 outputdir=None,
                                 nodoctype=False,
                                 noviewbox=False,
                                 inline_svg=False)
directive_options = {}


def relfn2path(env, filename):
    if filename.startswith('/') or filename.startswith(os.sep):
        relfn = filename[1:]
    else:
        path = env.doc2path(env.docname, base=None)
        relfn = os.path.join(os.path.dirname(path), filename)

    return relfn, os.path.join(env.srcdir, relfn)


def with_blockdiag(fn):
    @wraps(fn)
    def decorator(*args):
        with Application():
            return fn(*args)

    return decorator


def align(argument):
    align_values = ('left', 'center', 'right')
    return rst.directives.choice(argument, align_values)


def figwidth_value(argument):
    if argument.lower() == 'image':
        return 'image'
    else:
        return rst.directives.length_or_percentage_or_unitless(argument, 'px')


class BlockdiagDirectiveBase(rst.Directive):
    """ Directive to insert arbitrary dot markup. """
    name = "blockdiag"
    node_class = blockdiag_node

    has_content = True
    required_arguments = 0
    optional_arguments = 1
    final_argument_whitespace = False
    option_spec = {
        'alt': rst.directives.unchanged,
        'height': rst.directives.length_or_unitless,
        'width': rst.directives.length_or_percentage_or_unitless,
        'scale': rst.directives.percentage,
        'align': align,
        'caption': rst.directives.unchanged,
        'desctable': rst.directives.flag,
        'maxwidth': rst.directives.nonnegative_int,  # deprecated
        'name': rst.directives.unchanged,
        'class': rst.directives.class_option,
        'figwidth': figwidth_value,
        'figclass': rst.directives.class_option,
    }

    def run(self):
        if self.arguments:
            document = self.state.document
            if self.content:
                msg = ('%s directive cannot have both content and '
                       'a filename argument' % self.name)
                return [document.reporter.warning(msg, line=self.lineno)]

            try:
                filename = self.source_filename(self.arguments[0])
                fp = io.open(filename, 'r', encoding='utf-8-sig')
                try:
                    dotcode = fp.read()
                finally:
                    fp.close()
            except (IOError, OSError):
                msg = 'External %s file %r not found or reading it failed' % \
                      (self.name, filename)
                return [document.reporter.warning(msg, line=self.lineno)]
        else:
            dotcode = '\n'.join(self.content).strip()
            if not dotcode:
                msg = 'Ignoring "%s" directive without content.' % self.name
                return [self.state_machine.reporter.warning(msg,
                                                            line=self.lineno)]

        set_classes(self.options)
        node = self.node_class()
        results = [node]

        node['code'] = dotcode
        node['caption'] = self.options.pop('caption', None)
        node['options'] = self.options

        # for sphinxcontrib.* module (backward compatibility)
        node['alt'] = self.options.get('alt')

        # replace maxwidth option to width (backward compatibility)
        if 'maxwidth' in node['options']:
            node['options']['width'] = str(node['options']['maxwidth'])

            msg = ':maxwidth: option is deprecated. Use :width: option.'
            warning = self.state_machine.reporter.warning(msg,
                                                          line=self.lineno)
            results.append(warning)

        return results

    def source_filename(self, filename):
        if hasattr(self.state.document.settings, 'env'):
            env = self.state.document.settings.env
            rel_filename, filename = relfn2path(env, self.arguments[0])
            env.note_dependency(rel_filename)

        return filename


class BlockdiagDirective(BlockdiagDirectiveBase):
    processor = None  # backward compatibility for 1.4.0

    @with_blockdiag
    def run(self):
        figwidth = self.options.pop('figwidth', None)
        figclasses = self.options.pop('figclass', None)
        if self.options.get('caption'):
            align = self.options.pop('align', None)
        else:
            align = None

        results = super(BlockdiagDirective, self).run()

        node = results[0]
        if not isinstance(node, self.node_class):
            return results

        try:
            diagram = self.node2diagram(node)
        except Exception as e:
            raise self.warning(str(e))

        if 'desctable' in node['options']:
            results += self.description_tables(diagram)

        results[0] = self.node2image(node, diagram)
        self.add_name(results[0])

        if node.get('caption'):
            elem = nodes.Element()
            self.state.nested_parse(ViewList([node['caption']], source=''),
                                    self.content_offset, elem)
            caption_node = nodes.caption(elem[0].rawsource, '',
                                         *elem[0].children)
            caption_node.source = elem[0].source
            caption_node.line = elem[0].line

            fig = nodes.figure()
            fig += results[0]
            fig += caption_node

            if figwidth == 'image':
                width = self.get_actual_width(node, diagram)
                fig['width'] = str(width) + 'px'
            elif figwidth is not None:
                fig['width'] = figwidth
            if figclasses:
                fig['classes'] += figclasses
            if align:
                fig['align'] = align

            results[0] = fig

        return results

    @property
    def global_options(self):
        return directive_options

    def node2diagram(self, node):
        if hasattr(node, 'to_diagram'):
            return node.to_diagram()
        else:
            try:
                tree = self.processor.parser.parse_string(node['code'])
            except:
                code = '%s { %s }' % (self.name, node['code'])
                tree = self.processor.parser.parse_string(code)
                node['code'] = code  # replace if succeeded

            return self.processor.builder.ScreenNodeBuilder.build(tree)

    def get_actual_width(self, node, diagram):
        fontmap = self.create_fontmap()
        if hasattr(node, 'to_drawer'):
            drawer = node.to_drawer('SVG', None, fontmap,
                                    **self.global_options)
        else:
            drawer = self.processor.drawer.DiagramDraw('SVG', diagram,
                                                       None, fontmap=fontmap)

        return drawer.pagesize()[0]

    def node2image(self, node, diagram):
        _format = self.global_options['format'].lower()
        if _format == 'svg' and self.global_options['inline_svg'] is True:
            return self.node2image_inline_svg(node, diagram)

        filename = self.image_filename(node)
        fontmap = self.create_fontmap()
        if hasattr(node, 'to_drawer'):
            drawer = node.to_drawer(_format, filename, fontmap,
                                    **self.global_options)
        else:
            drawer = self.processor.drawer.DiagramDraw(_format, diagram,
                                                       filename,
                                                       fontmap=fontmap,
                                                       **self.global_options)

        if not os.path.isfile(filename):
            drawer.draw()
            drawer.save()

        return nodes.image(uri=filename, **node['options'])

    def node2image_inline_svg(self, node, diagram):
        fontmap = self.create_fontmap()
        if hasattr(node, 'to_drawer'):
            drawer = node.to_drawer('SVG', None, fontmap,
                                    **self.global_options)
        else:
            drawer = self.processor.drawer.DiagramDraw('svg', diagram,
                                                       None, fontmap=fontmap,
                                                       **self.global_options)
        drawer.draw()

        size = drawer.pagesize().resize(**node['options']).to_integer_point()
        content = drawer.save(size)

        return nodes.raw('', content, format='html')

    def create_fontmap(self):
        Options = namedtuple('Options', 'font fontmap')
        fontpath = self.global_options['fontpath']
        if isinstance(fontpath, (list, tuple)):
            options = Options(fontpath, None)
        elif isinstance(fontpath, string_types):
            options = Options([fontpath], None)
        else:
            options = Options([], None)

        return create_fontmap(options)

    def image_filename(self, node, prefix='', ext='png'):
        if hasattr(node, 'get_path'):
            return node.get_path(**self.global_options)
        else:
            options = dict(node['options'])
            options.update(font=self.global_options['fontpath'],
                           antialias=self.global_options['antialias'])
            hashseed = (node['code'] + str(options)).encode('utf-8')
            hashed = sha1(hashseed).hexdigest()

            _format = self.global_options['format'].lower()
            outputdir = self.global_options['outputdir']
            filename = "%s%s-%s.%s" % (self.name, prefix, hashed, _format)
            if outputdir:
                filename = os.path.join(outputdir, filename)

            return filename

    def description_tables(self, diagram):
        tables = []
        desctable = self.node_description_table(diagram)
        if desctable:
            tables.append(desctable)

        desctable = self.edge_description_table(diagram)
        if desctable:
            tables.append(desctable)

        return tables

    def node_description_table(self, diagram):
        nodes = diagram.traverse_nodes()
        klass = diagram._DiagramNode

        widths = [25] + [50] * (len(klass.desctable) - 1)
        headers = [klass.attrname[n] for n in klass.desctable]

        def node_number(node):
            try:
                return int(node[0])
            except (TypeError, ValueError):
                return 65535

        descriptions = [n.to_desctable() for n in nodes if n.drawable]
        descriptions.sort(key=node_number)

        for i in reversed(range(len(headers))):
            if any(desc[i] for desc in descriptions):
                pass
            else:
                widths.pop(i)
                headers.pop(i)
                for desc in descriptions:
                    desc.pop(i)

        if len(headers) == 1:
            return None
        else:
            return self._description_table(descriptions, widths, headers)

    def edge_description_table(self, diagram):
        edges = diagram.traverse_edges()

        widths = [25, 50]
        headers = ['Name', 'Description']
        descriptions = [e.to_desctable() for e in edges if e.style != 'none']

        if any(desc[1] for desc in descriptions):
            return self._description_table(descriptions, widths, headers)
        else:
            return None

    def _description_table(self, descriptions, widths, headers):
        # generate table-root
        tgroup = nodes.tgroup(cols=len(widths))
        for width in widths:
            tgroup += nodes.colspec(colwidth=width)
        table = nodes.table()
        table += tgroup

        # generate table-header
        thead = nodes.thead()
        row = nodes.row()
        for header in headers:
            entry = nodes.entry()
            entry += nodes.paragraph(text=header)
            row += entry
        thead += row
        tgroup += thead

        # generate table-body
        tbody = nodes.tbody()
        for desc in descriptions:
            row = nodes.row()
            for attr in desc:
                entry = nodes.entry()
                if not isinstance(attr, string_types):
                    attr = str(attr)
                self.state.nested_parse(ViewList([attr], source=attr),
                                        0, entry)
                row += entry
            tbody += row
        tgroup += tbody

        return table


def setup(**kwargs):
    global directive_options, directive_options_default

    for key, value in directive_options_default.items():
        directive_options[key] = kwargs.get(key, value)

    rst.directives.register_directive("blockdiag", BlockdiagDirective)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from blockdiag.noderenderer import install_renderer
from blockdiag.noderenderer.base import NodeShape


class NoneShape(NodeShape):
    def __init__(self, node, metrics=None):
        super(NoneShape, self).__init__(node, metrics)

        p = metrics.cell(node).center
        self.connectors = [p, p, p, p]

    def render_label(self, drawer, **kwargs):
        pass

    def render_shape(self, drawer, _, **kwargs):
        pass


def setup(self):
    install_renderer('none', NoneShape)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import io
import os
import sys
from docutils import nodes
from docutils.core import publish_doctree
from docutils.parsers.rst import directives as docutils
from blockdiag.utils.rst import directives
from blockdiag.utils.rst.nodes import blockdiag as blockdiag_node
from blockdiag.tests.utils import capture_stderr, TemporaryDirectory

if sys.version_info < (2, 7):
    import unittest2 as unittest
else:
    import unittest


class TestRstDirectives(unittest.TestCase):
    def setUp(self):
        docutils.register_directive('blockdiag',
                                    directives.BlockdiagDirectiveBase)
        self._tmpdir = TemporaryDirectory()

    def tearDown(self):
        if 'blockdiag' in docutils._directives:
            del docutils._directives['blockdiag']

        self._tmpdir.clean()

    @capture_stderr
    def test_without_args(self):
        text = ".. blockdiag::"
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.system_message, type(doctree[0]))

    def test_block(self):
        text = (".. blockdiag::\n"
                "\n"
                "   { A -> B }")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(blockdiag_node, type(doctree[0]))
        self.assertEqual('{ A -> B }', doctree[0]['code'])
        self.assertEqual(None, doctree[0]['alt'])
        self.assertEqual({}, doctree[0]['options'])

    @capture_stderr
    def test_emptyblock(self):
        text = ".. blockdiag::\n\n   \n"
        text = (".. blockdiag::\n"
                "\n"
                "   ")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.system_message, type(doctree[0]))

    def test_filename(self):
        dirname = os.path.dirname(__file__)
        filename = os.path.join(dirname, '../diagrams/node_attribute.diag')
        text = ".. blockdiag:: %s" % filename
        doctree = publish_doctree(text)

        self.assertEqual(1, len(doctree))
        self.assertEqual(blockdiag_node, type(doctree[0]))
        self.assertEqual(io.open(filename, encoding='utf-8-sig').read(),
                         doctree[0]['code'])
        self.assertEqual(None, doctree[0]['alt'])
        self.assertEqual({}, doctree[0]['options'])

    @capture_stderr
    def test_filename_not_exists(self):
        text = ".. blockdiag:: unknown.diag"
        doctree = publish_doctree(text)
        self.assertEqual(nodes.system_message, type(doctree[0]))

    @capture_stderr
    def test_both_block_and_filename(self):
        text = (".. blockdiag:: unknown.diag\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(nodes.system_message, type(doctree[0]))

    def test_full_options(self):
        text = (".. blockdiag::\n"
                "   :alt: hello world\n"
                "   :align: center\n"
                "   :desctable:\n"
                "   :width: 200\n"
                "   :height: 100\n"
                "   :scale: 50%\n"
                "   :name: foo\n"
                "   :class: bar\n"
                "   :figwidth: 400\n"
                "   :figclass: baz\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(1, len(doctree))
        self.assertEqual(blockdiag_node, type(doctree[0]))
        self.assertEqual('A -> B', doctree[0]['code'])
        self.assertEqual('hello world', doctree[0]['alt'])
        self.assertEqual('center', doctree[0]['options']['align'])
        self.assertEqual(None, doctree[0]['options']['desctable'])
        self.assertEqual('200', doctree[0]['options']['width'])
        self.assertEqual('100', doctree[0]['options']['height'])
        self.assertEqual(50, doctree[0]['options']['scale'])
        self.assertEqual('hello world', doctree[0]['options']['alt'])
        self.assertEqual('foo', doctree[0]['options']['name'])
        self.assertEqual(['bar'], doctree[0]['options']['classes'])
        self.assertEqual('400px', doctree[0]['options']['figwidth'])
        self.assertEqual(['baz'], doctree[0]['options']['figclass'])

    @capture_stderr
    def test_maxwidth_option(self):
        text = (".. blockdiag::\n"
                "   :maxwidth: 200\n"
                "\n"
                "   A -> B")
        doctree = publish_doctree(text)
        self.assertEqual(2, len(doctree))
        self.assertEqual(blockdiag_node, type(doctree[0]))
        self.assertEqual('A -> B', doctree[0]['code'])
        self.assertEqual('200', doctree[0]['options']['width'])
        self.assertEqual(nodes.system_message, type(doctree[1]))



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import math
from blockdiag.utils import XY

DIVISION = 1000.0
CYCLE = 10


def _angles(du, a, b, start, end):
    phi = (start / 180.0) * math.pi
    while phi <= (end / 180.0) * math.pi:
        yield phi
        phi += du / math.sqrt((a * math.sin(phi)) ** 2 +
                              (b * math.cos(phi)) ** 2)


def _coordinates(du, a, b, start, end):
    for angle in _angles(du, a, b, start, end):
        yield (a * math.cos(angle), b * math.sin(angle))


def endpoints(du, a, b, start, end):
    pt1 = next(iter(_coordinates(du, a, b, start, start + 1)))
    pt2 = next(iter(_coordinates(du, a, b, end, end + 1)))

    return [XY(*pt1), XY(*pt2)]


def dots(box, cycle, start=0, end=360):
    # calcrate rendering pattern from cycle
    base = 0
    rendered = []
    for index in range(0, len(cycle), 2):
        i, j = cycle[index:index + 2]
        for n in range(base * 2, (base + i) * 2):
            rendered.append(n)
        base += i + j

    a = float(box.width) / 2
    b = float(box.height) / 2
    du = 1
    _max = sum(cycle) * 2
    center = box.center
    for i, coord in enumerate(_coordinates(du, a, b, start, end)):
        if i % _max in rendered:
            yield XY(center.x + coord[0], center.y + coord[1])



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from blockdiag.noderenderer import install_renderer
from blockdiag.noderenderer.base import NodeShape
from blockdiag.utils import XY, Box


class EndPoint(NodeShape):
    def __init__(self, node, metrics=None):
        super(EndPoint, self).__init__(node, metrics)

        m = metrics.cell(node)

        self.radius = metrics.cellsize
        self.center = m.center
        self.textbox = Box(m.top.x, m.top.y, m.right.x, m.right.y)
        self.textalign = 'left'
        self.connectors = [XY(self.center.x, self.center.y - self.radius),
                           XY(self.center.x + self.radius, self.center.y),
                           XY(self.center.x, self.center.y + self.radius),
                           XY(self.center.x - self.radius, self.center.y)]

    def render_shape(self, drawer, _, **kwargs):
        fill = kwargs.get('fill')

        # draw outer circle
        r = self.radius
        box = Box(self.center.x - r, self.center.y - r,
                  self.center.x + r, self.center.y + r)
        if kwargs.get('shadow'):
            box = self.shift_shadow(box)
            if kwargs.get('style') == 'blur':
                drawer.ellipse(box, fill=fill, outline=fill,
                               filter='transp-blur')
            else:
                drawer.ellipse(box, fill=fill, outline=fill)
        else:
            drawer.ellipse(box, fill='white', outline=self.node.linecolor,
                           style=self.node.style)

        # draw inner circle
        box = Box(self.center.x - r / 2, self.center.y - r / 2,
                  self.center.x + r / 2, self.center.y + r / 2)
        if not kwargs.get('shadow'):
            if self.node.color == self.node.basecolor:
                color = self.node.linecolor
            else:
                color = self.node.color

            drawer.ellipse(box, fill=color, outline=self.node.linecolor,
                           style=self.node.style)


def setup(self):
    install_renderer('endpoint', EndPoint)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division
import copy
from collections import defaultdict
from blockdiag import noderenderer
from blockdiag.elements import DiagramNode
from blockdiag.utils import Box, Size, XY
from blockdiag.utils.fontmap import FontInfo, FontMap

cellsize = 8


class EdgeLines(object):
    def __init__(self):
        self.xy = None
        self.stroking = False
        self.polylines = []

    def moveTo(self, x, y=None):
        self.stroking = False
        if y is None:
            self.xy = x
        else:
            self.xy = XY(x, y)

    def lineTo(self, x, y=None):
        if y is None:
            elem = x
        else:
            elem = XY(x, y)

        if self.stroking is False:
            self.stroking = True
            polyline = []
            if self.xy:
                polyline.append(self.xy)
            self.polylines.append(polyline)

        if len(self.polylines[-1]) > 0:
            if self.polylines[-1][-1] == elem:
                return

        self.polylines[-1].append(elem)

    def lines(self):
        lines = []
        for line in self.polylines:
            start = line[0]
            for elem in list(line[1:]):
                lines.append((start, elem))
                start = elem

        return lines


class AutoScaler(object):
    def __init__(self, subject, scale_ratio):
        self.subject = subject
        self.scale_ratio = scale_ratio

    def __getattr__(self, name):
        ratio = self.scale_ratio
        return self.scale(getattr(self.subject, name), ratio)

    def __getitem__(self, name):
        ratio = self.scale_ratio
        return self.scale(self.subject[name], ratio)

    @classmethod
    def scale(cls, value, ratio):
        if not callable(value):
            return cls._scale(value, ratio)
        else:
            def _(*args, **kwargs):
                ret = value(*args, **kwargs)
                return cls._scale(ret, ratio)

            return _

    @classmethod
    def _scale(cls, value, ratio):
        if ratio == 1:
            return value

        klass = value.__class__
        if klass == XY:
            ret = XY(value.x * ratio, value.y * ratio)
        elif klass == Size:
            ret = Size(value.width * ratio, value.height * ratio)
        elif klass == Box:
            ret = Box(value[0] * ratio, value[1] * ratio,
                      value[2] * ratio, value[3] * ratio)
        elif klass == tuple:
            ret = tuple([cls.scale(x, ratio) for x in value])
        elif klass == list:
            ret = [cls.scale(x, ratio) for x in value]
        elif klass == EdgeLines:
            ret = EdgeLines()
            ret.polylines = cls.scale(value.polylines, ratio)
        elif klass == FontInfo:
            ret = FontInfo(value.familyname, value.path, value.size * ratio)
        elif klass == int:
            ret = value * ratio
        elif klass == str:
            ret = value
        else:
            ret = cls(value, ratio)

        return ret

    @property
    def original_metrics(self):
        return self.subject


class DiagramMetrics(object):
    cellsize = cellsize
    edge_layout = 'normal'
    node_padding = 4
    line_spacing = 2
    shadow_offset = XY(3, 6)
    page_margin = XY(0, 0)
    page_padding = [0, 0, 0, 0]
    node_width = cellsize * 16
    node_height = cellsize * 5
    span_width = cellsize * 8
    span_height = cellsize * 5

    def __init__(self, diagram, drawer=None, fontmap=None):
        self.drawer = drawer

        if diagram.node_width is not None:
            self.node_width = diagram.node_width

        if diagram.node_height is not None:
            self.node_height = diagram.node_height

        if diagram.span_width is not None:
            self.span_width = diagram.span_width

        if diagram.span_height is not None:
            self.span_height = diagram.span_height

        if fontmap is not None:
            self.fontmap = fontmap
        else:
            self.fontmap = FontMap()

        if diagram.page_padding is not None:
            self.page_padding = diagram.page_padding

        if diagram.edge_layout is not None:
            self.edge_layout = diagram.edge_layout

        # setup spreadsheet
        sheet = self.spreadsheet = SpreadSheetMetrics(self)
        nodes = [n for n in diagram.traverse_nodes() if n.drawable]

        node_width = self.node_width
        for x in range(diagram.colwidth):
            widths = [n.width for n in nodes if n.xy.x == x]
            if widths:
                width = max(n or node_width for n in widths)
                sheet.set_node_width(x, width)

        node_height = self.node_height
        for y in range(diagram.colheight):
            heights = [n.height for n in nodes if n.xy.y == y]
            if heights:
                height = max(n or node_height for n in heights)
                sheet.set_node_height(y, height)

    @property
    def original_metrics(self):
        return self

    def shift(self, x, y):
        metrics = copy.copy(self)
        metrics.spreadsheet = copy.copy(self.spreadsheet)
        metrics.spreadsheet.metrics = metrics
        metrics.page_margin = XY(x, y)

        return metrics

    def textsize(self, string, font=None, width=65535):
        return self.drawer.textsize(string, font, maxwidth=width)

    def node(self, node):
        renderer = noderenderer.get(node.shape)

        if hasattr(renderer, 'render'):
            return renderer(node, self)
        else:
            return self.cell(node)

    def cell(self, node, use_padding=True):
        return self.spreadsheet.node(node, use_padding)

    def group(self, group):
        return self.spreadsheet.node(group, use_padding=False)

    def edge(self, edge):
        if self.edge_layout == 'flowchart':
            if edge.node1.group.orientation == 'landscape':
                return FlowchartLandscapeEdgeMetrics(edge, self)
            else:
                return FlowchartPortraitEdgeMetrics(edge, self)
        else:
            if edge.node1.group.orientation == 'landscape':
                return LandscapeEdgeMetrics(edge, self)
            else:
                return PortraitEdgeMetrics(edge, self)

    def font_for(self, element):
        return self.fontmap.find(element)

    def pagesize(self, width, height):
        return self.spreadsheet.pagesize(width, height)


class SubMetrics(object):
    def __getattr__(self, name):
        # avoid recursion-error on Python 2.6
        if 'metrics' not in self.__dict__:
            raise AttributeError()

        return getattr(self.metrics, name)


class SpreadSheetMetrics(SubMetrics):
    def __init__(self, metrics):
        self.metrics = metrics
        self.node_width = defaultdict(lambda: metrics.node_width)
        self.node_height = defaultdict(lambda: metrics.node_height)
        self.span_width = defaultdict(lambda: metrics.span_width)
        self.span_height = defaultdict(lambda: metrics.span_height)

    def set_node_width(self, x, width):
        if (width is not None and 0 < width and
           (x not in self.node_width or self.node_width[x] < width)):
            self.node_width[x] = width

    def set_node_height(self, y, height):
        if (height is not None and 0 < height and
           (y not in self.node_height or self.node_height[y] < height)):
            self.node_height[y] = height

    def set_span_width(self, x, width):
        if (width is not None and 0 < width and
           (x not in self.span_width or self.span_width[x] < width)):
            self.span_width[x] = width

    def add_span_width(self, x, width):
        self.span_width[x] += width

    def set_span_height(self, y, height):
        if (height is not None and 0 < height and
           (y not in self.span_height or self.span_height[y] < height)):
            self.span_height[y] = height

    def add_span_height(self, y, height):
        self.span_height[y] += height

    def node(self, node, use_padding=True):
        x1, y1 = self._node_topleft(node, use_padding)
        x2, y2 = self._node_bottomright(node, use_padding)

        return NodeMetrics(self.metrics, x1, y1, x2, y2)

    def _node_topleft(self, node, use_padding=True):
        x, y = node.xy
        margin = self.page_margin
        padding = self.page_padding

        node_width = sum(self.node_width[i] for i in range(x))
        node_height = sum(self.node_height[i] for i in range(y))
        span_width = sum(self.span_width[i] for i in range(x + 1))
        span_height = sum(self.span_height[i] for i in range(y + 1))

        if use_padding:
            width = node.width or self.metrics.node_width
            xdiff = (self.node_width[x] - width) // 2
            if xdiff < 0:
                xdiff = 0

            height = node.height or self.metrics.node_height
            ydiff = (self.node_height[y] - height) // 2
            if ydiff < 0:
                ydiff = 0
        else:
            xdiff = 0
            ydiff = 0

        x1 = margin.x + padding[3] + node_width + span_width + xdiff
        y1 = margin.y + padding[0] + node_height + span_height + ydiff

        return XY(x1, y1)

    def _node_bottomright(self, node, use_padding=True):
        x = node.xy.x + node.colwidth - 1
        y = node.xy.y + node.colheight - 1
        margin = self.page_margin
        padding = self.page_padding

        node_width = sum(self.node_width[i] for i in range(x + 1))
        node_height = sum(self.node_height[i] for i in range(y + 1))
        span_width = sum(self.span_width[i] for i in range(x + 1))
        span_height = sum(self.span_height[i] for i in range(y + 1))

        if use_padding:
            width = node.width or self.metrics.node_width
            xdiff = (self.node_width[x] - width) // 2
            if xdiff < 0:
                xdiff = 0

            height = node.height or self.metrics.node_height
            ydiff = (self.node_height[y] - height) // 2
            if ydiff < 0:
                ydiff = 0
        else:
            xdiff = 0
            ydiff = 0

        x2 = margin.x + padding[3] + node_width + span_width - xdiff
        y2 = margin.y + padding[0] + node_height + span_height - ydiff

        return XY(x2, y2)

    def pagesize(self, width, height):
        margin = self.metrics.page_margin
        padding = self.metrics.page_padding

        dummy = DiagramNode(None)
        dummy.xy = XY(width - 1, height - 1)
        x, y = self._node_bottomright(dummy, use_padding=False)
        x_span = self.span_width[width]
        y_span = self.span_height[height]
        return Size(x + margin.x + padding[1] + x_span,
                    y + margin.y + padding[2] + y_span)


class NodeMetrics(SubMetrics):
    def __init__(self, metrics, x1, y1, x2, y2):
        self.metrics = metrics
        self._box = Box(x1, y1, x2, y2)

    def __getattr__(self, name):
        if hasattr(self._box, name):
            return getattr(self._box, name)
        else:
            return getattr(self.metrics, name)

    def __getitem__(self, key):
        return self.box[key]

    @property
    def box(self):
        return self._box

    @property
    def marginbox(self):
        return Box(self._box.x1 - self.span_width // 8,
                   self._box.y1 - self.span_height // 4,
                   self._box.x2 + self.span_width // 8,
                   self._box.y2 + self.span_height // 4)

    @property
    def corebox(self):
        return Box(self._box.x1 + self.node_padding,
                   self._box.y1 + self.node_padding,
                   self._box.x2 - self.node_padding * 2,
                   self._box.y2 - self.node_padding * 2)

    @property
    def grouplabelbox(self):
        return Box(self._box.x1, self._box.y1 - self.span_height // 2,
                   self._box.x2, self._box.y1)


class EdgeMetrics(SubMetrics):
    def __init__(self, edge, metrics):
        self.metrics = metrics
        self.edge = edge

    @property
    def headshapes(self):
        pass

    @property
    def _shaft(self):
        pass

    @property
    def heads(self):
        heads = []
        head1, head2 = self.headshapes

        if head1:
            heads.append(self._head(self.edge.node1, head1))

        if head2:
            heads.append(self._head(self.edge.node2, head2))

        return heads

    def _head(self, node, direct):
        head = []
        cell = self.cellsize
        node = self.node(node)

        if direct == 'up':
            xy = node.bottom
            head.append(XY(xy.x, xy.y + 1))
            head.append(XY(xy.x - cell // 2, xy.y + cell))
            head.append(XY(xy.x, xy.y + cell * 2))
            head.append(XY(xy.x + cell // 2, xy.y + cell))
            head.append(XY(xy.x, xy.y + 1))
        elif direct == 'down':
            xy = node.top
            head.append(XY(xy.x, xy.y - 1))
            head.append(XY(xy.x - cell // 2, xy.y - cell))
            head.append(XY(xy.x, xy.y - cell * 2))
            head.append(XY(xy.x + cell // 2, xy.y - cell))
            head.append(XY(xy.x, xy.y - 1))
        elif direct == 'right':
            xy = node.left
            head.append(XY(xy.x - 1, xy.y))
            head.append(XY(xy.x - cell, xy.y - cell // 2))
            head.append(XY(xy.x - cell * 2, xy.y))
            head.append(XY(xy.x - cell, xy.y + cell // 2))
            head.append(XY(xy.x - 1, xy.y))
        elif direct == 'left':
            xy = node.right
            head.append(XY(xy.x + 1, xy.y))
            head.append(XY(xy.x + cell, xy.y - cell // 2))
            head.append(XY(xy.x + cell * 2, xy.y))
            head.append(XY(xy.x + cell, xy.y + cell // 2))
            head.append(XY(xy.x + 1, xy.y))
        elif direct == 'rup':
            xy = node.bottom
            head.append(XY(xy.x, xy.y + cell))
            head.append(XY(xy.x - cell, xy.y + 1))
            head.append(XY(xy.x, xy.y + 1 * 2))
            head.append(XY(xy.x + cell, xy.y + 1))
            head.append(XY(xy.x, xy.y + cell))
        elif direct == 'rdown':
            xy = node.top
            head.append(XY(xy.x, xy.y - cell))
            head.append(XY(xy.x - cell, xy.y - 1))
            head.append(XY(xy.x, xy.y - 1 * 2))
            head.append(XY(xy.x + cell, xy.y - 1))
            head.append(XY(xy.x, xy.y - cell))
        elif direct == 'rright':
            xy = node.left
            head.append(XY(xy.x - cell, xy.y))
            head.append(XY(xy.x - 1, xy.y - cell))
            head.append(XY(xy.x - 1 * 2, xy.y))
            head.append(XY(xy.x - 1, xy.y + cell))
            head.append(XY(xy.x - cell, xy.y))
        elif direct == 'rleft':
            xy = node.right
            head.append(XY(xy.x + cell, xy.y))
            head.append(XY(xy.x + 1, xy.y - cell))
            head.append(XY(xy.x + 1 * 2, xy.y))
            head.append(XY(xy.x + 1, xy.y + cell))
            head.append(XY(xy.x + cell, xy.y))

        if self.edge.hstyle not in ('composition', 'aggregation'):
            head.pop(2)

        return head

    @property
    def shaft(self):
        cell = self.cellsize
        lines = self._shaft
        head1, head2 = self.headshapes

        if head1:
            pt = lines.polylines[0].pop(0)
            if head1 == 'up':
                lines.polylines[0].insert(0, XY(pt.x, pt.y + cell))
            elif head1 == 'right':
                lines.polylines[0].insert(0, XY(pt.x - cell, pt.y))
            elif head1 == 'left':
                lines.polylines[0].insert(0, XY(pt.x + cell, pt.y))
            elif head1 == 'down':
                lines.polylines[0].insert(0, XY(pt.x, pt.y - cell))
            elif head1 == 'rup':
                lines.polylines[0].insert(0, XY(pt.x, pt.y + cell))
            elif head1 == 'rright':
                lines.polylines[0].insert(0, XY(pt.x - cell, pt.y))
            elif head1 == 'rleft':
                lines.polylines[0].insert(0, XY(pt.x + cell, pt.y))
            elif head1 == 'rdown':
                lines.polylines[0].insert(0, XY(pt.x, pt.y - cell))

        if head2:
            pt = lines.polylines[-1].pop()
            if head2 == 'up':
                lines.polylines[-1].append(XY(pt.x, pt.y + cell))
            elif head2 == 'right':
                lines.polylines[-1].append(XY(pt.x - cell, pt.y))
            elif head2 == 'left':
                lines.polylines[-1].append(XY(pt.x + cell, pt.y))
            elif head2 == 'down':
                lines.polylines[-1].append(XY(pt.x, pt.y - cell))
            elif head2 == 'rup':
                lines.polylines[-1].append(XY(pt.x, pt.y + cell))
            elif head2 == 'rright':
                lines.polylines[-1].append(XY(pt.x - cell, pt.y))
            elif head2 == 'rleft':
                lines.polylines[-1].append(XY(pt.x + cell, pt.y))
            elif head2 == 'rdown':
                lines.polylines[-1].append(XY(pt.x, pt.y - cell))

        return lines

    @property
    def labelbox(self):
        pass


class LandscapeEdgeMetrics(EdgeMetrics):
    @property
    def headshapes(self):
        heads = []
        _dir = self.edge.direction

        if self.edge.dir in ('back', 'both'):
            if _dir in ('left-up', 'left', 'same',
                        'right-up', 'right', 'right-down'):
                heads.append('left')
            elif _dir == 'up':
                if self.edge.skipped:
                    heads.append('left')
                else:
                    heads.append('down')
            elif _dir in ('left-down', 'down'):
                if self.edge.skipped:
                    heads.append('left')
                else:
                    heads.append('up')

            if self.edge.hstyle in ('manyone', 'manymany'):
                heads[-1] = 'r' + heads[-1]
        else:
            heads.append(None)

        if self.edge.dir in ('forward', 'both'):
            if _dir in ('right-up', 'right', 'right-down'):
                heads.append('right')
            elif _dir == 'up':
                heads.append('up')
            elif _dir in ('left-up', 'left', 'left-down', 'down', 'same'):
                heads.append('down')

            if self.edge.hstyle in ('onemany', 'manymany'):
                heads[-1] = 'r' + heads[-1]
        else:
            heads.append(None)

        return heads

    @property
    def _shaft(self):
        span = XY(self.span_width, self.span_height)
        _dir = self.edge.direction

        node1 = self.node(self.edge.node1)
        cell1 = self.cell(self.edge.node1, use_padding=False)
        node2 = self.node(self.edge.node2)
        cell2 = self.cell(self.edge.node2, use_padding=False)

        shaft = EdgeLines()
        if _dir == 'right':
            shaft.moveTo(node1.right)

            if self.edge.skipped:
                shaft.lineTo(cell1.right.x + span.x // 2, cell1.right.y)
                shaft.lineTo(cell1.right.x + span.x // 2,
                             cell1.bottomright.y + span.y // 2)
                shaft.lineTo(cell2.left.x - span.x // 4,
                             cell2.bottomright.y + span.y // 2)
                shaft.lineTo(cell2.left.x - span.x // 4, cell2.left.y)

            shaft.lineTo(node2.left)

        elif _dir == 'right-up':
            shaft.moveTo(node1.right)

            if self.edge.skipped:
                shaft.lineTo(cell1.right.x + span.x // 2, cell1.right.y)
                shaft.lineTo(cell1.right.x + span.x // 2,
                             cell2.bottomleft.y + span.y // 2)
                shaft.lineTo(cell2.left.x - span.x // 4,
                             cell2.bottomleft.y + span.y // 2)
                shaft.lineTo(cell2.left.x - span.x // 4, cell2.left.y)
            else:
                shaft.lineTo(cell2.left.x - span.x // 4, cell1.right.y)
                shaft.lineTo(cell2.left.x - span.x // 4, cell2.left.y)

            shaft.lineTo(node2.left)

        elif _dir == 'right-down':
            shaft.moveTo(node1.right)
            shaft.lineTo(cell1.right.x + span.x // 2, cell1.right.y)

            if self.edge.skipped:
                shaft.lineTo(cell1.right.x + span.x // 2,
                             cell2.topleft.y - span.y // 2)
                shaft.lineTo(cell2.left.x - span.x // 4,
                             cell2.topleft.y - span.y // 2)
                shaft.lineTo(cell2.left.x - span.x // 4, cell2.left.y)
            else:
                shaft.lineTo(cell1.right.x + span.x // 2, cell2.left.y)

            shaft.lineTo(node2.left)

        elif _dir == 'up':
            if self.edge.skipped:
                shaft.moveTo(node1.right)
                shaft.lineTo(cell1.right.x + span.x // 4, cell1.right.y)
                shaft.lineTo(cell1.right.x + span.x // 4,
                             cell2.bottom.y + span.y // 2)
                shaft.lineTo(cell2.bottom.x, cell2.bottom.y + span.y // 2)
            else:
                shaft.moveTo(node1.top)

            shaft.lineTo(node2.bottom)

        elif _dir in ('left-up', 'left', 'same'):
            shaft.moveTo(node1.right)
            shaft.lineTo(cell1.right.x + span.x // 4, cell1.right.y)
            shaft.lineTo(cell1.right.x + span.x // 4,
                         cell2.top.y - span.y // 2 + span.y // 8)
            shaft.lineTo(cell2.top.x,
                         cell2.top.y - span.y // 2 + span.y // 8)
            shaft.lineTo(node2.top)

        elif _dir == 'left-down':
            if self.edge.skipped:
                shaft.moveTo(node1.right)
                shaft.lineTo(cell1.right.x + span.x // 2, cell1.right.y)
                shaft.lineTo(cell1.right.x + span.x // 2,
                             cell2.top.y - span.y // 2)
                shaft.lineTo(cell2.top.x, cell2.top.y - span.y // 2)
            else:
                shaft.moveTo(node1.bottom)
                shaft.lineTo(cell1.bottom.x,
                             cell2.top.y - span.y // 2)
                shaft.lineTo(cell2.top.x, cell2.top.y - span.y // 2)

            shaft.lineTo(node2.top)

        elif _dir == 'down':
            if self.edge.skipped:
                shaft.moveTo(node1.right)
                shaft.lineTo(cell1.right.x + span.x // 2, cell1.right.y)
                shaft.lineTo(cell1.right.x + span.x // 2,
                             cell2.top.y - span.y // 2 + span.y // 8)
                shaft.lineTo(cell2.top.x,
                             cell2.top.y - span.y // 2 + span.y // 8)
            else:
                shaft.moveTo(node1.bottom)

            shaft.lineTo(node2.top)

        return shaft

    @property
    def labelbox(self):
        span = XY(self.span_width, self.span_height)
        node = XY(self.node_width, self.node_height)

        _dir = self.edge.direction
        node1 = self.cell(self.edge.node1, use_padding=False)
        node2 = self.cell(self.edge.node2, use_padding=False)

        if _dir == 'right':
            if self.edge.skipped:
                box = Box(node1.bottomright.x + span.x,
                          node1.bottomright.y,
                          node2.bottomleft.x - span.x,
                          node2.bottomleft.y + span.y // 2)
            else:
                box = Box(node1.topright.x, node1.topright.y - span.y // 8,
                          node2.left.x, node2.left.y - span.y // 8)

        elif _dir == 'right-up':
            box = Box(node2.left.x - span.x, node1.top.y - node.y // 2,
                      node2.bottomleft.x, node1.top.y)

        elif _dir == 'right-down':
            box = Box(node1.right.x, node2.topleft.y - span.y // 8,
                      node1.right.x + span.x, node2.left.y - span.y // 8)

        elif _dir in ('up', 'left-up', 'left', 'same'):
            if self.edge.node2.xy.y < self.edge.node1.xy.y:
                box = Box(node1.topright.x - span.x // 2 + span.x // 4,
                          node1.topright.y - span.y // 2,
                          node1.topright.x + span.x // 2 + span.x // 4,
                          node1.topright.y)
            else:
                box = Box(node1.top.x + span.x // 4,
                          node1.top.y - span.y,
                          node1.topright.x + span.x // 4,
                          node1.topright.y - span.y // 2)

        elif _dir in ('left-down', 'down'):
            box = Box(node2.top.x + span.x // 4,
                      node2.top.y - span.y,
                      node2.topright.x + span.x // 4,
                      node2.topright.y - span.y // 2)

        # shrink box
        box = Box(box[0] + span.x // 8, box[1],
                  box[2] - span.x // 8, box[3])

        return box


class PortraitEdgeMetrics(EdgeMetrics):
    @property
    def headshapes(self):
        heads = []
        _dir = self.edge.direction

        if self.edge.dir in ('back', 'both'):
            if _dir == 'right':
                if self.edge.skipped:
                    heads.append('up')
                else:
                    heads.append('left')
            elif _dir in ('up', 'right-up', 'same'):
                heads.append('up')
            elif _dir in ('left-up', 'left'):
                heads.append('left')
            elif _dir in ('left-down', 'down', 'right-down'):
                if self.edge.skipped:
                    heads.append('left')
                else:
                    heads.append('up')

            if self.edge.hstyle in ('manyone', 'manymany'):
                heads[-1] = 'r' + heads[-1]
        else:
            heads.append(None)

        if self.edge.dir in ('forward', 'both'):
            if _dir == 'right':
                if self.edge.skipped:
                    heads.append('down')
                else:
                    heads.append('right')
            elif _dir in ('up', 'right-up', 'same'):
                heads.append('down')
            elif _dir in ('left-up', 'left',
                          'left-down', 'down', 'right-down'):
                heads.append('down')

            if self.edge.hstyle in ('onemany', 'manymany'):
                heads[-1] = 'r' + heads[-1]
        else:
            heads.append(None)

        return heads

    @property
    def _shaft(self):
        span = XY(self.span_width, self.span_height)
        _dir = self.edge.direction

        node1 = self.node(self.edge.node1)
        cell1 = self.cell(self.edge.node1, use_padding=False)
        node2 = self.node(self.edge.node2)
        cell2 = self.cell(self.edge.node2, use_padding=False)

        shaft = EdgeLines()
        if _dir in ('up', 'right-up', 'same', 'right'):
            if _dir == 'right' and not self.edge.skipped:
                shaft.moveTo(node1.right)
                shaft.lineTo(node2.left)
            else:
                shaft.moveTo(node1.bottom)
                shaft.lineTo(cell1.bottom.x, cell1.bottom.y + span.y // 2)
                shaft.lineTo(cell2.right.x + span.x // 4,
                             cell1.bottom.y + span.y // 2)
                shaft.lineTo(cell2.right.x + span.x // 4,
                             cell2.top.y - span.y // 2 + span.y // 8)
                shaft.lineTo(cell2.top.x,
                             cell2.top.y - span.y // 2 + span.y // 8)
                shaft.lineTo(node2.top)

        elif _dir == 'right-down':
            shaft.moveTo(node1.bottom)
            shaft.lineTo(cell1.bottom.x, cell1.bottom.y + span.y // 2)

            if self.edge.skipped:
                shaft.lineTo(cell2.left.x - span.x // 2,
                             cell1.bottom.y + span.y // 2)
                shaft.lineTo(cell2.topleft.x - span.x // 2,
                             cell2.topleft.y - span.y // 2)
                shaft.lineTo(cell2.top.x, cell2.top.y - span.y // 2)
            else:
                shaft.lineTo(cell2.top.x, cell1.bottom.y + span.y // 2)

            shaft.lineTo(node2.top)

        elif _dir in ('left-up', 'left', 'same'):
            shaft.moveTo(node1.right)
            shaft.lineTo(cell1.right.x + span.x // 4, cell1.right.y)
            shaft.lineTo(cell1.right.x + span.x // 4,
                         cell2.top.y - span.y // 2 + span.y // 8)
            shaft.lineTo(cell2.top.x,
                         cell2.top.y - span.y // 2 + span.y // 8)
            shaft.lineTo(node2.top)

        elif _dir == 'left-down':
            shaft.moveTo(node1.bottom)

            if self.edge.skipped:
                shaft.lineTo(cell1.bottom.x, cell1.bottom.y + span.y // 2)
                shaft.lineTo(cell2.right.x + span.x // 2,
                             cell1.bottom.y + span.y // 2)
                shaft.lineTo(cell2.right.x + span.x // 2,
                             cell2.top.y - span.y // 2)
            else:
                shaft.lineTo(cell1.bottom.x, cell2.top.y - span.y // 2)

            shaft.lineTo(cell2.top.x, cell2.top.y - span.y // 2)
            shaft.lineTo(node2.top)

        elif _dir == 'down':
            shaft.moveTo(node1.bottom)

            if self.edge.skipped:
                shaft.lineTo(cell1.bottom.x, cell1.bottom.y + span.y // 2)
                shaft.lineTo(cell1.right.x + span.x // 2,
                             cell1.bottom.y + span.y // 2)
                shaft.lineTo(cell2.right.x + span.x // 2,
                             cell2.top.y - span.y // 2)
                shaft.lineTo(cell2.top.x, cell2.top.y - span.y // 2)

            shaft.lineTo(node2.top)

        return shaft

    @property
    def labelbox(self):
        span = XY(self.span_width, self.span_height)

        _dir = self.edge.direction
        node1 = self.cell(self.edge.node1, use_padding=False)
        node2 = self.cell(self.edge.node2, use_padding=False)

        if _dir == 'right':
            if self.edge.skipped:
                box = Box(node1.bottomright.x + span.x,
                          node1.bottomright.y,
                          node2.bottomleft.x - span.x,
                          node2.bottomleft.y + span.y // 2)
            else:
                box = Box(node1.topright.x, node1.topright.y - span.y // 8,
                          node2.left.x, node2.left.y - span.y // 8)

        elif _dir == 'right-up':
            box = Box(node2.left.x - span.x, node2.left.y,
                      node2.bottomleft.x, node2.bottomleft.y)

        elif _dir == 'right-down':
            box = Box(node2.topleft.x, node2.topleft.y - span.y // 2,
                      node2.top.x, node2.top.y)

        elif _dir in ('up', 'left-up', 'left', 'same'):
            if self.edge.node2.xy.y < self.edge.node1.xy.y:
                box = Box(node1.topright.x - span.x // 2 + span.x // 4,
                          node1.topright.y - span.y // 2,
                          node1.topright.x + span.x // 2 + span.x // 4,
                          node1.topright.y)
            else:
                box = Box(node1.top.x + span.x // 4,
                          node1.top.y - span.y,
                          node1.topright.x + span.x // 4,
                          node1.topright.y - span.y // 2)

        elif _dir == 'down':
            box = Box(node2.top.x + span.x // 4,
                      node2.top.y - span.y // 2,
                      node2.topright.x + span.x // 4,
                      node2.topright.y)

        elif _dir == 'left-down':
            box = Box(node1.bottomleft.x, node1.bottomleft.y,
                      node1.bottom.x, node1.bottom.y + span.y // 2)

        # shrink box
        box = Box(box[0] + span.x // 8, box[1],
                  box[2] - span.x // 8, box[3])

        return box


class FlowchartLandscapeEdgeMetrics(LandscapeEdgeMetrics):
    @property
    def headshapes(self):
        heads = []

        if self.edge.direction == 'right-down':
            if self.edge.dir in ('back', 'both'):
                if self.edge.hstyle in ('manyone', 'manymany'):
                    heads.append('rup')
                else:
                    heads.append('up')
            else:
                heads.append(None)

            if self.edge.dir in ('forward', 'both'):
                if self.edge.hstyle in ('onemany', 'manymany'):
                    heads.append('rright')
                else:
                    heads.append('right')
            else:
                heads.append(None)
        else:
            heads = super(FlowchartLandscapeEdgeMetrics, self).headshapes

        return heads

    @property
    def _shaft(self):
        if self.edge.direction == 'right-down':
            span = XY(self.span_width, self.span_height)
            node1 = self.node(self.edge.node1)
            cell1 = self.cell(self.edge.node1, use_padding=False)
            node2 = self.node(self.edge.node2)
            cell2 = self.cell(self.edge.node2, use_padding=False)

            shaft = EdgeLines()
            shaft.moveTo(node1.bottom)

            if self.edge.skipped:
                shaft.lineTo(cell1.bottom.x, cell1.bottom.y + span.y // 2)
                shaft.lineTo(cell2.left.x - span.x // 4,
                             cell1.bottom.y + span.y // 2)
                shaft.lineTo(cell2.left.x - span.x // 4, cell2.left.y)
            else:
                shaft.lineTo(cell1.bottom.x, cell2.left.y)

            shaft.lineTo(node2.left)
        else:
            shaft = super(FlowchartLandscapeEdgeMetrics, self)._shaft

        return shaft

    @property
    def labelbox(self):
        _dir = self.edge.direction
        if _dir == 'right':
            span = XY(self.span_width, self.span_height)
            cell1 = self.cell(self.edge.node1, use_padding=False)
            cell2 = self.cell(self.edge.node2, use_padding=False)

            if self.edge.skipped:
                box = Box(cell1.bottom.x, cell1.bottom.y,
                          cell1.bottomright.x,
                          cell1.bottomright.y + span.y // 2)
            else:
                box = Box(cell1.bottom.x, cell2.left.y - span.y // 2,
                          cell1.bottom.x, cell2.left.y)
        else:
            box = super(FlowchartLandscapeEdgeMetrics, self).labelbox

        return box


class FlowchartPortraitEdgeMetrics(PortraitEdgeMetrics):
    @property
    def headshapes(self):
        heads = []

        if self.edge.direction == 'right-down':
            if self.edge.dir in ('back', 'both'):
                if self.edge.hstyle in ('manyone', 'manymany'):
                    heads.append('left')
                else:
                    heads.append('left')
            else:
                heads.append(None)

            if self.edge.dir in ('forward', 'both'):
                if self.edge.dir in ('onemany', 'manymany'):
                    heads.append('rdown')
                else:
                    heads.append('down')
            else:
                heads.append(None)
        else:
            heads = super(FlowchartPortraitEdgeMetrics, self).headshapes

        return heads

    @property
    def _shaft(self):
        if self.edge.direction == 'right-down':
            span = XY(self.span_width, self.span_height)
            node1 = self.node(self.edge.node1)
            cell1 = self.cell(self.edge.node1, use_padding=False)
            node2 = self.node(self.edge.node2)
            cell2 = self.cell(self.edge.node2, use_padding=False)

            shaft = EdgeLines()
            shaft.moveTo(node1.right)

            if self.edge.skipped:
                shaft.lineTo(cell1.right.x + span.x * 3 // 4,
                             cell1.right.y)
                shaft.lineTo(cell1.right.x + span.x * 3 // 4,
                             cell2.topleft.y - span.y // 2)
                shaft.lineTo(cell2.top.x, cell2.top.y - span.y // 2)
            else:
                shaft.lineTo(cell2.top.x, cell1.right.y)

            shaft.lineTo(node2.top)
        else:
            shaft = super(FlowchartPortraitEdgeMetrics, self)._shaft

        return shaft

    @property
    def labelbox(self):
        _dir = self.edge.direction
        span = XY(self.span_width, self.span_height)
        cell1 = self.cell(self.edge.node1, use_padding=False)
        cell2 = self.cell(self.edge.node2, use_padding=False)

        if _dir == 'down':
            box = Box(cell2.topleft.x, cell2.top.y - span.y // 2,
                      cell2.top.x, cell2.top.y)
        elif _dir == 'right':
            if self.edge.skipped:
                box = Box(cell1.bottom.x, cell1.bottom.y,
                          cell1.bottomright.x,
                          cell1.bottomright.y + span.y // 2)
            else:
                box = Box(cell1.bottom.x, cell2.left.y - span.y // 2,
                          cell1.bottom.x, cell2.left.y)
        else:
            box = super(FlowchartPortraitEdgeMetrics, self).labelbox

        return box



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import print_function

import sys
from blockdiag.parser import parse_string, ParseException
from blockdiag.parser import Diagram, Group, Statements, Node, Edge

if sys.version_info < (2, 7):
    import unittest2 as unittest
else:
    import unittest


class TestParser(unittest.TestCase):
    def test_basic(self):
        # basic digram
        code = """
               diagram test {
                  A -> B -> C, D;
               }
               """

        tree = parse_string(code)
        self.assertIsInstance(tree, Diagram)

    def test_without_diagram_id(self):
        code = """
               diagram {
                  A -> B -> C, D;
               }
               """
        tree = parse_string(code)
        self.assertIsInstance(tree, Diagram)

        code = """
               {
                  A -> B -> C, D;
               }
               """
        tree = parse_string(code)
        self.assertIsInstance(tree, Diagram)

    def test_empty_diagram(self):
        code = """
               diagram {
               }
               """
        tree = parse_string(code)
        self.assertIsInstance(tree, Diagram)

        code = """
               {
               }
               """
        tree = parse_string(code)
        self.assertIsInstance(tree, Diagram)

    def test_diagram_includes_nodes(self):
        code = """
               diagram {
                 A;
                 B [label = "foobar"];
                 C [color = "red"];
               }
               """
        tree = parse_string(code)
        self.assertIsInstance(tree, Diagram)
        self.assertEqual(3, len(tree.stmts))
        self.assertIsInstance(tree.stmts[0], Statements)
        self.assertIsInstance(tree.stmts[0].stmts[0], Node)
        self.assertIsInstance(tree.stmts[1], Statements)
        self.assertIsInstance(tree.stmts[1].stmts[0], Node)
        self.assertIsInstance(tree.stmts[2], Statements)
        self.assertIsInstance(tree.stmts[2].stmts[0], Node)

    def test_diagram_includes_edges(self):
        code = """
               diagram {
                 A -> B -> C;
               }
               """
        tree = parse_string(code)
        self.assertIsInstance(tree, Diagram)
        self.assertEqual(1, len(tree.stmts))
        self.assertIsInstance(tree.stmts[0], Statements)
        self.assertEqual(2, len(tree.stmts[0].stmts))
        self.assertIsInstance(tree.stmts[0].stmts[0], Edge)
        self.assertIsInstance(tree.stmts[0].stmts[1], Edge)

        code = """
               diagram {
                 A -> B -> C [style = dotted];
                 D -> E, F;
               }
               """
        tree = parse_string(code)
        self.assertIsInstance(tree, Diagram)
        self.assertEqual(2, len(tree.stmts))
        self.assertIsInstance(tree.stmts[0], Statements)
        self.assertEqual(2, len(tree.stmts[0].stmts))
        self.assertIsInstance(tree.stmts[0].stmts[0], Edge)
        self.assertIsInstance(tree.stmts[0].stmts[1], Edge)
        self.assertIsInstance(tree.stmts[1], Statements)
        self.assertEqual(1, len(tree.stmts[1].stmts))
        self.assertIsInstance(tree.stmts[1].stmts[0], Edge)

    def test_diagram_includes_groups(self):
        code = """
               diagram {
                 group {
                   A; B;
                 }
                 group {
                   C -> D;
                 }
               }
               """
        tree = parse_string(code)
        self.assertIsInstance(tree, Diagram)
        self.assertEqual(2, len(tree.stmts))

        self.assertIsInstance(tree.stmts[0], Group)
        self.assertEqual(2, len(tree.stmts[0].stmts))
        self.assertIsInstance(tree.stmts[0].stmts[0], Statements)
        self.assertIsInstance(tree.stmts[0].stmts[0].stmts[0], Node)
        self.assertIsInstance(tree.stmts[0].stmts[1], Statements)
        self.assertIsInstance(tree.stmts[0].stmts[1].stmts[0], Node)

        self.assertIsInstance(tree.stmts[1], Group)
        self.assertEqual(1, len(tree.stmts[1].stmts))
        self.assertIsInstance(tree.stmts[1].stmts[0], Statements)
        self.assertIsInstance(tree.stmts[1].stmts[0].stmts[0], Edge)

    def test_diagram_includes_diagram_attributes(self):
        code = """
               diagram {
                 fontsize = 12;
                 node_width = 80;
               }
               """
        tree = parse_string(code)
        self.assertIsInstance(tree, Diagram)
        self.assertEqual(2, len(tree.stmts))

    def test_parenthesis_ness(self):
        with self.assertRaises(ParseException):
            code = ""
            parse_string(code)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import sys
from blockdiag.imagedraw.textfolder import splitlabel
from blockdiag.imagedraw.textfolder import splittext
from blockdiag.imagedraw.textfolder import truncate_text
from blockdiag.utils import Size
from blockdiag.utils.compat import u

if sys.version_info < (2, 7):
    import unittest2 as unittest
else:
    import unittest


CHAR_WIDTH = 14
CHAR_HEIGHT = 10


class Metrics(object):
    def textsize(self, text):
        length = len(text)
        return Size(CHAR_WIDTH * length, CHAR_HEIGHT)


class TestTextFolder(unittest.TestCase):
    def test_splitlabel(self):
        # single line text
        text = "abc"
        self.assertEqual(['abc'], list(splitlabel(text)))

        # text include \n (as char a.k.a. \x5c)
        text = "abc\ndef"
        self.assertEqual(['abc', 'def'], list(splitlabel(text)))

        # text include \n (as mac yensign a.k.a. \xa5)
        text = "abc\xa5ndef"
        self.assertEqual(['abc', 'def'], list(splitlabel(text)))

        # text includes \n (as text)
        text = "abc\\ndef"
        self.assertEqual(['abc', 'def'], list(splitlabel(text)))

        # text includes escaped \n
        text = "abc\\\\ndef"
        self.assertEqual(['abc\\ndef'], list(splitlabel(text)))

        # text includes escaped \n (\x5c and mac yensign mixed)
        if sys.version_info[0] == 2:
            text = u("abc\xa5\\\\ndef")
        else:
            text = u("abc\xa5\\ndef")
        self.assertEqual(['abc\\ndef'], list(splitlabel(text)))

        # text include \n and spaces
        text = " abc \n def "
        self.assertEqual(['abc', 'def'], list(splitlabel(text)))

        # text starts empty line
        text = " \nabc\ndef"
        self.assertEqual(['abc', 'def'], list(splitlabel(text)))

        # text starts empty line with \n (as text)
        text = " \\nabc\\ndef"
        self.assertEqual(['', 'abc', 'def'], list(splitlabel(text)))

    def test_splittext_width(self):
        metrics = Metrics()

        # just fit text
        text = "abc"
        ret = splittext(metrics, text, CHAR_WIDTH * 3)
        self.assertEqual(['abc'], ret)

        # text should be folded (once)
        text = "abcdef"
        ret = splittext(metrics, text, CHAR_WIDTH * 3)
        self.assertEqual(['abc', 'def'], ret)

        # text should be folded (twice)
        text = "abcdefghi"
        ret = splittext(metrics, text, CHAR_WIDTH * 3)
        self.assertEqual(['abc', 'def', 'ghi'], ret)

        # empty text
        text = ""
        ret = splittext(metrics, text, CHAR_WIDTH * 3)
        self.assertEqual([' '], ret)

    def test_truncate_text(self):
        metrics = Metrics()

        # truncated
        text = "abcdef"
        ret = truncate_text(metrics, text, CHAR_WIDTH * 8)
        self.assertEqual("abcd ...", ret)

        # truncated
        text = "abcdef"
        ret = truncate_text(metrics, text, CHAR_WIDTH * 5)
        self.assertEqual("a ...", ret)

        # not truncated (too short)
        text = "abcdef"
        ret = truncate_text(metrics, text, CHAR_WIDTH * 4)
        self.assertEqual("abcdef", ret)
